from typing import Optional, Union, List

from quickstats.components import AbstractObject, ExtendedModel, ExtendedMinimizer
from quickstats.utils.common_utils import get_class_that_defined_method
import ROOT

class AnalysisObject(AbstractObject):
    def __init__(self, filename:Optional[str]=None, poi_name:Optional[str]=None, 
                 data_name:str='combData', binned_likelihood:bool=True, 
                 fix_param:str='', profile_param:str='', ws_name:Optional[str]=None, 
                 mc_name:Optional[str]=None, snapshot_name:Optional[Union[List[str], str]]=None,
                 minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', precision:float=0.001, 
                 eps:float=1.0, strategy:int=1, print_level:int=-1, timer:bool=False,
                 num_cpu:int=1, offset:bool=True, optimize:int=2, eigen:bool=False,
                 fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1, 
                 max_iters:int=-1, constrain_nuis:bool=True, batch_mode:bool=False,
                 int_bin_precision:float=-1., verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)    
        self.model = None
        self.minimizer = None
        
        if filename is not None:
            self.setup_model(filename, data_name, binned_likelihood, ws_name, mc_name, 
                             snapshot_name, fix_cache, fix_multi, verbosity=verbosity)
            self.set_poi(poi_name)
            self.setup_parameters(fix_param, profile_param)
            self.setup_minimizer(minimizer_type, minimizer_algo, precision, eps, strategy, num_cpu,
                                 offset, optimize, eigen, max_calls, max_iters, print_level,
                                 timer, constrain_nuis, batch_mode, int_bin_precision,
                                 verbosity=verbosity)
    # black magic
    def _inherit_init(self, init_func, **kwargs):
        import inspect
        this_parameters = list(inspect.signature(AnalysisObject.__init__).parameters)
        if "self" in this_parameters:
            this_parameters.remove("self")
        that_parameters = list(inspect.signature(init_func).parameters)
        is_calling_this = set(this_parameters) == set(that_parameters)
        if is_calling_this:
            init_func(**kwargs)
        else:
            that_kwargs = {k:v for k,v in kwargs.items() if k in that_parameters}
            this_kwargs = {k:v for k,v in kwargs.items() if k not in that_parameters}
            init_func(config=this_kwargs, **that_kwargs)

    def set_poi(self, poi_name:Optional[str]=None):
        self.poi = self.model.get_poi(poi_name)
        if poi_name is not None:
            self.stdout.info('INFO: POI set to "{}"'.format(poi_name))            
        
    def setup_model(self, filename, data_name='combData', binned_likelihood:bool=True, 
                    ws_name:Optional[str]=None, mc_name:Optional[str]=None, 
                    snapshot_name:Optional[str]=None, fix_cache:bool=True, 
                    fix_multi:bool=True, verbosity:Optional[Union[int, str]]="INFO"):
        model = ExtendedModel(fname=filename, ws_name=ws_name, mc_name=mc_name, data_name=data_name, 
                              snapshot_name=snapshot_name, binned_likelihood=binned_likelihood,
                              fix_cache=fix_cache, fix_multi=fix_multi, verbosity=verbosity)
        self.model = model
    
    def setup_parameters(self, fix_param:str='', profile_param:str=''):
        if not self.model:
            raise RuntimeError('uninitialized analysis object')
        if fix_param:
            self.model.fix_parameters(fix_param)
        if profile_param:
            self.model.profile_parameters(profile_param)
            
    def setup_minimizer(self, minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', 
                        precision:float=0.001, eps:float=1.0, strategy:int=0, 
                        num_cpu:int=1, offset:bool=True, optimize:bool=True, eigen:bool=False,
                        max_calls:int=-1, max_iters:int=-1, print_level:int=-1, timer:bool=False,
                        constrain_nuis:bool=True, batch_mode:bool=False, int_bin_precision:float=-1.,
                        verbosity:Optional[Union[int, str]]="INFO"):
        
        minimizer = ExtendedMinimizer("Minimizer", self.model.pdf, self.model.data, verbosity=verbosity)
        
        nll_commands = [ROOT.RooFit.NumCPU(num_cpu, 3), 
                        ROOT.RooFit.Offset(offset)]
        if (constrain_nuis and self.model.nuisance_parameters):
            #nll_commands.append(ROOT.RooFit.Constrain(self.model.nuisance_parameters))
            nll_commands.append(ROOT.RooFit.GlobalObservables(self.model.global_observables))
            nll_commands.append(ROOT.RooFit.ConditionalObservables(self.model.model_config.GetConditionalObservables()))
            #nll_commands.append(ROOT.RooFit.ExternalConstraints(ROOT.RooArgSet()))
        if hasattr(ROOT.RooFit, "BatchMode"):
            nll_commands.append(ROOT.RooFit.BatchMode(batch_mode))
        if hasattr(ROOT.RooFit, "IntegrateBins"):
            nll_commands.append(ROOT.RooFit.IntegrateBins(int_bin_precision))
        minimizer_options = {
            'minimizer_type'   : minimizer_type,
            'minimizer_algo'   : minimizer_algo,
            'default_strategy' : strategy,
            'opt_const'        : optimize,
            'precision'        : precision,
            'eps'              : eps,
            'eigen'            : eigen,
            'max_calls'        : max_calls,
            'max_iters'        : max_iters,
            'print_level'      : print_level,
            'timer'            : timer
        }

        minimizer.configure(nll_commands=nll_commands, **minimizer_options)
        
        self.minimizer = minimizer
        self.minimizer_options = minimizer_options
        self.nll_commands      = nll_commands
    
    def set_data(self, data_name:str='combData'):
        data = self.model.workspace.data(data_name)
        if not data:
            raise RuntimeError(f"workspace does not contain the dataset \"{data_name}\"")
        self.minimizer.set_data(data)
        self.model._data = data
        
    def load_snapshot(self, snapshot_name:Optional[str]=None):
        if snapshot_name is not None:
            self.model.workspace.loadSnapshot(snapshot_name)
        
    def save(self, fname:str, recreate:bool=True):
        self.model.save(fname, recreate=recreate)
        self.stdout.info(f"INFO: Saved workspace file as `{fname}`")


from functools import partial

class semistaticmethod(object):
    def __init__(self, callable):
        self.f = callable
    def __get__(self, obj, type=None):
        if (obj is None) and (type is not None):
            return partial(self.f, type)
        if (obj is not None):
            return partial(self.f, obj)
        return self.f
    @property
    def __func__(self):
        return self.f


from typing import Union, Optional, Dict, List

import pandas as pd

import ROOT

def copy_attributes(source:"ROOT.RooAbsArg", target:"ROOT.RooAbsArg"):
    if source is target:
        return
    for attrib in source.attributes():
        target.setAttribute(attrib)
    for attrib in source.stringAttributes():
        target.setStringAttribute(attrib.first, attrib.second)
        
def get_variable_attributes(variable:ROOT.RooRealVar):
    attributes = {
        'name' : variable.GetName(),
        'value': variable.getVal(),
        'error': variable.getError(),
        'min'  : variable.getMin(),
        'max'  : variable.getMax(),
        'is_constant': variable.isConstant()
    }
    return attributes

def variable_collection_to_dataframe(var_collection:Union[list, ROOT.RooArgSet, ROOT.RooArgSet]):
    data = []
    for variable in var_collection:
        attributes = get_variable_attributes(variable)
        data.append(attributes)
    df = pd.DataFrame(data)
    return df
        
def construct_categorized_pdf_dataset(pdf:"ROOT.RooAbsPdf", dataset:"ROOT.RooDataSet", 
                               workspace:"ROOT.RooWorkspace", label:str,
                               category_name:str="adhocCat"):
    # make sure pdf is not already a simultaneous pdf
    assert pdf.ClassName() != "RooSimultaneous"
    cat = ROOT.RooCategory(category_name, category_name)
    cat.defineType(label)
    pdf_dict = {label: pdf}
    pdf_map = ExtendedModel.get_object_map(pdf_dict, "RooAbsPdf")
    dataset_dict = {label: dataset}
    dataset_map = ExtendedModel.get_object_map(data_dict, "RooDataSet")
    sim_pdf = ROOT.RooSimultaneous(pdf.GetName(), pdf.GetName(), pdf_map, cat)
    obs_and_weight = dataset.get()
    weight_var = workspace.var("weightVar")
    if not weitht_var:
        raise RuntimeError("workspace does not contain the variable `weightVar`")
    obs_and_weight.add(weight_var)
    indexed_dataset = ROOT.RooDataSet(dataset.GetName(), dataset.GetName(), obs_and_weight, 
                                      ROOT.RooFit.Index(cat),
                                      ROOT.RooFit.Import(dataset_map),
                                      ROOT.RooFit.WeightVar(weight_var))
    return sim_pdf, indexed_dataset


def factorize_pdf(observables:"ROOT.RooArgSet", pdf:"ROOT.RooAbsPdf", constraints:"ROOT.RooArgSet"):
    pdf_class = pdf.ClassName()   
    if pdf_class == "RooProdPdf":
        new_factors = ROOT.RooArgList()
        new_owned = ROOT.RooArgSet()
        pdf_list = pdf.pdfList()
        need_new = False
        for i in range(len(pdf_list)):
            pdf_i = pdf_list[i]
            new_pdf = factorize_pdf(observables, pdf_i, constraints)
            if new_pdf == 0:
                need_new = True
                continue
            if new_pdf is not pdf_i:
                need_new = True
                new_owned.add(new_pdf)
            new_factors.add(new_pdf)
        if not need_new:
            return pdf
        elif len(new_factors) == 0:
            return 0
        elif len(new_factors) == 1:
            clone_pdf = new_factors.first().Clone("{}_obsOnly".format(pdf.GetName()))
            copy_attributes(pdf, clone_pdf)
            return clone_pdf
        factorized_pdf = ROOT.RooProdPdf("{}_obsOnly".format(pdf.GetName()), "", new_factors)
        factorized_pdf.addOwnedComponents(new_owned)
        copy_attributes(pdf, factorized_pdf)
        return factorized_pdf
    elif pdf_class == "RooSimultaneous":
        cat = pdf.indexCat().Clone()
        n_bins = cat.numBins("")
        factorized_pdfs = []
        new_owned = ROOT.RooArgSet()
        need_new = False
        for i in range(n_bins):
            cat.setBin(i)
            pdf_i = pdf.getPdf(cat.getLabel())
            new_pdf = factorize_pdf(observables, pdf_i, constraints)
            factorized_pdfs.append(new_pdf)     
            if new_pdf == 0:
                raise RuntimeError("channel `{}` factorized to 0".format(cat.getLabel()))
            if new_pdf is not pdf_i:
                need_new = True
                new_owned.add(new_pdf)
                # this can be removed after version 6.28
                ROOT.SetOwnership(new_pdf, False)
        factorized_pdf = pdf
        if need_new:
            factorized_pdf = ROOT.RooSimultaneous("{}_obsOnly".format(pdf.GetName()), "", pdf.indexCat())
            for i in range(n_bins):
                cat.setBin(i)
                new_pdf = factorized_pdfs[i]
                if new_pdf:
                    factorized_pdf.addPdf(new_pdf, cat.getLabel())
            factorized_pdf.addOwnedComponents(new_owned)
        # has to delete persistent object
        cat.Delete()
        copy_attributes(pdf, factorized_pdf)
        return factorized_pdf         
    elif pdf.dependsOn(observables):
        return pdf
    else:
        if not constraints.contains(pdf):
            constraints.add(pdf)
        return 0
    
    
def rebuild_simultaneous_pdf(observables:"ROOT.RooArgSet", sim_pdf:"ROOT.RooSimultaneous"):
    assert sim_pdf.ClassName() == "RooSimultaneous"
    constraints = ROOT.RooArgList()
    cat = sim_pdf.indexCat().Clone()
    n_bins = cat.numBins("")
    factorized_pdfs = []
    new_owned = ROOT.RooArgSet()
    for i in range(n_bins):
        cat.setBin(i)
        pdf_i = sim_pdf.getPdf(cat.getLabel())
        if pdf_i == 0:
            factorized_pdfs.append(0)
            continue
        new_pdf = factorize_pdf(observables, pdf_i, constraints)     
        factorized_pdfs.append(new_pdf)
        if new_pdf == 0:
            continue
        if new_pdf is not pdf_i:
            new_owned.add(new_pdf)
            # this can be removed after version 6.28
            ROOT.SetOwnership(new_pdf, False)
    rebuilt_pdf = ROOT.RooSimultaneous("{}_reloaded".format(sim_pdf.GetName()), "", sim_pdf.indexCat())
    for i in range(n_bins):
        cat.setBin(i)
        new_pdf = factorized_pdfs[i]
        if new_pdf:
            if constraints.getSize() > 0:
                all_factors = ROOT.RooArgList(constraints)
                all_factors.add(new_pdf)
                newer_pdf = ROOT.RooProdPdf("{}_plus_constr".format(new_pdf.GetName()), "",
                                           all_factors)
                rebuilt_pdf.addPdf(newer_pdf, cat.getLabel())
                copy_attributes(new_pdf, newer_pdf)
                new_owned.add(newer_pdf)
                # this can be removed after version 6.28
                ROOT.SetOwnership(newer_pdf, False)
            else:
                rebuilt_pdf.addPdf(new_pdf, cat.getLabel())
    rebuilt_pdf.addOwnedComponents(new_owned)
    copy_attributes(sim_pdf, rebuilt_pdf)
    return rebuilt_pdf



def print_object(obj, indent=0, spacer="  ", prefix="-", show_address=False):
    if show_address:
        print(f"{spacer*indent}{prefix}{obj.GetName()}({obj.ClassName()} @ {hex(id(obj))})")
    else:
        print(f"{spacer*indent}{prefix}{obj.GetName()}({obj.ClassName()})")
        
def print_pdf_structure(pdf:"ROOT.RooAbsPdf", level:int=0, show_address=False):
    print_object(pdf, level, show_address=show_address)
    class_name = pdf.ClassName()
    if class_name == "RooSimultaneous":
        cat = pdf.indexCat().Clone()
        n_bins = cat.numBins("")
        for i in range(n_bins):
            cat.setBin(i)
            pdf_i = pdf.getPdf(cat.getLabel())
            print_pdf_structure(pdf_i, level+1, show_address=show_address)
    elif class_name == "RooProdPdf":
        pdf_list = pdf.pdfList()
        for pdf_i in pdf_list:
            print_pdf_structure(pdf_i, level+1, show_address=show_address)


import os
import click

from quickstats.components.processors import RooProcessor

@click.command(name='process_rfile')
@click.option('-i', '--input_file', 'filename', required=True, 
              help='Input ROOT file to process.')
@click.option('-c', '--config', 'config_path', required=True,
              help='Path to the processor card.')
@click.option('--multithread/--no-multirhread', default=True, show_default=True,
              help='Enable implicit multi-threading.')
@click.option('-t', '--tag', default=None,
              help='Include `tag` as a global variable.')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def process_rfile(filename, config_path, multithread, tag, verbosity):
    """
    Process a ROOT file based on RDataFrame routines.
    """
    rprocessor = RooProcessor(config_path, multithread=multithread, verbosity=verbosity)
    rprocessor.global_variables['tag'] = tag
    rprocessor.run(filename)


from typing import Dict, Optional, Union, List
import pandas as pd

from quickstats.plots import MultiDataPlot
from quickstats.plots.template import create_transform
from quickstats.utils.common_utils import combine_dict

class Likelihood1DPlot(MultiDataPlot):
    
    CONFIG = {
        'sigma_values': (1, 4),
        'sigma_line_styles':{
            'color': 'gray',
            'linestyle': '--'
        }
    }
    def __init__(self, data_map:Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 label_map:Optional[Dict]=None,
                 styles_map:Optional[Dict]=None,
                 color_cycler=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None):
        super().__init__(data_map=data_map,
                         label_map=label_map,
                         styles_map=styles_map,
                         color_cycler=color_cycler,
                         styles=styles,
                         analysis_label_options=analysis_label_options)
        self.config = combine_dict(self.CONFIG)
        
    def draw_single_data(self, ax, data:pd.DataFrame, 
                         styles:Optional[Dict]=None,
                         label:Optional[str]=None,
                         xattrib:str='mu', yattrib:str='qmu'):
        x = data[xattrib]
        y = data[yattrib]
        draw_styles = combine_dict(self.styles['plot'], styles)
        handle = ax.plot(x, y, **draw_styles, label=label)
        return handle[0]
    
    def draw(self, xattrib:str='mu', yattrib:str='qmu', xlabel:Optional[str]=None, 
             ylabel:Optional[str]="$-2\Delta ln(L)$",
             ymax:float=7, xmin:Optional[float]=None, xmax:Optional[float]=None,
             draw_sigma_line:bool=True):
        ax = self.draw_frame()
        if isinstance(self.data_map, pd.DataFrame):
            self.draw_single_data(ax, self.data_map, self.styles_map,
                                  xattrib=xattrib, yattrib=yattrib)
        elif isinstance(self.data_map, dict):
            if self.styles_map is None:
                styles_map = {k:None for k in self.data_map}
            else:
                styles_map = self.styles_map
            if self.label_map is None:
                label_map = {k:k for k in data_map}
            else:
                label_map = self.label_map
            handles = {}
            for key in self.data_map:
                data = self.data_map[key]
                styles = styles_map.get(key, None)
                label = label_map.get(key, "")
                handle = self.draw_single_data(ax, data, styles, 
                                               label=label,
                                               xattrib=xattrib,
                                               yattrib=yattrib)
                handles[key] = handle
            self.update_legend_handles(handles)
        else:
            raise ValueError("invalid data format")
            
        if draw_sigma_line:
            transform = create_transform(transform_x="axis", transform_y="data")
            sigma_line_styles = self.config['sigma_line_styles']
            sigma_values = self.config['sigma_values']
            ax.hlines(sigma_values, xmin=0, xmax=1, transform=transform,
                      **sigma_line_styles)
            
        handles, labels = self.get_legend_handles_labels()
        ax.legend(handles, labels, **self.styles['legend'])
        ax.set_ylim([0, ymax])
        ax.set_xlim([xmin, xmax])
        
        self.draw_axis_labels(ax, xlabel=xlabel, ylabel=ylabel)
        
        return ax


import click

import fnmatch
from quickstats.components import RooInspector
    
@click.command(name='inspect_rfile')
@click.option('-i', '--file_expr', required=True, help='File name expression '
              '(supports wild card, multiple files separated by commas).')
@click.option('-t', '--tree_name', required=True, help='Tree name.')
@click.option('-o', '--output_file', default=None, help='Export output to text file. If None, no output is saved.')
@click.option('--include', 'include_patterns', default=None, 
              help='Match variable names with given patterns (separated by commas).')
@click.option('--exclude', 'exclude_patterns', default=None,
              help='Exclude variable names with given patterns (separated by commas).')
@click.option('-f','--filter', 'filter_expr', default=None, show_default=True,
              help='Apply a filter to the events.')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def inspect_rfile(file_expr, tree_name, filter_expr=None, output_file=None,
                  include_patterns=None, exclude_patterns=None, verbosity="INFO"):
    '''
        Inspect root files
    '''
    rinspector = RooInspector(tree_name, file_expr, filter_expr=filter_expr, verbosity=verbosity)
    rinspector.print_summary(include_patterns=include_patterns,
                             exclude_patterns=exclude_patterns,
                             save_as=output_file)


from typing import Optional, Union, Dict
import os
import sys
import copy
import time
import json
import inspect
import datetime
import functools
import collections.abc
from concurrent.futures  import ProcessPoolExecutor

import numpy as np

def timely_info(green_text, normal_text, adjust=40):
    print(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), '\033[92m[INFO]\033[0m', '\033[92m{}\033[0m'.format(green_text).rjust(40, ' '), normal_text)

def get_cpu_count():
    return os.cpu_count()

def parallel_run(func, *iterables, max_workers):

    with ProcessPoolExecutor(max_workers) as executor:
        result = executor.map(func, *iterables)

    return [i for i in result]

def execute_multi_tasks(func, *iterables, parallel):
    if parallel == 0:
        result = []
        for args in zip(*iterables):
            result.append(func(*args))
        return result
    else:
        if parallel == -1:
            max_workers = get_cpu_count()
        else:
            max_workers = parallel
        return parallel_run(func, *iterables, max_workers=max_workers)


def timer(func):
    """Print the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        print('INFO: All jobs have finished. Total time taken: {:.3f} s'.format(run_time))
        return value
    return wrapper_timer

def stdout_print(msg):
    sys.__stdout__.write(msg + '\n')
    
def redirect_stdout(logfile_path):
    import ROOT
    sys.stdout = open(logfile_path, 'w')
    ROOT.gSystem.RedirectOutput(logfile_path)

def restore_stdout():
    import ROOT
    if sys.stdout != sys.__stdout__:
        sys.stdout.close()
    sys.stdout = sys.__stdout__
    ROOT.gROOT.ProcessLine('gSystem->RedirectOutput(0);')

def redirect_stdout_test(func):
    """Redirect stdout to a log file"""
    import ROOT
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        logfile_path = kwargs.get('logfile_path', None)
        if logfile_path is not None:
            sys.stdout = open(logfile_path, 'w')
            ROOT.gSystem.RedirectOutput(logfile_path)
            value = func(*args, **kwargs)
            sys.stdout.close()
            sys.stdout = sys.__stdout__
            ROOT.gROOT.ProcessLine('gSystem->RedirectOutput(0);')
            return value
        else:
            return func(*args, **kwargs)
    return wrapper_timer

def json_load(fp, *args, **kwargs):
    try:
        data = json.load(fp, *args, **kwargs)
    except:
        raise RuntimeError(f"broken json input: {fp}")
    return data


def parse_config(source:Optional[Union[Dict, str]]=None):
    if source is None:
        return {}
    elif isinstance(source, str):
        with open(source, 'r') as f:
                config = json.load(f)
        return config
    elif isinstance(source, dict):
        return source
    else:
        raise ValueError("invalid config input")
        
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)
    
def update_nested_dict(d:Dict, u:Dict):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update_nested_dict(d.get(k, {}), v)
        else:
            d[k] = v
    return d

def combine_dict(d:Dict, u:Optional[Dict]=None):
    d_copy = copy.deepcopy(d)
    if u is None:
        return d_copy
    else:
        u_copy = copy.deepcopy(u)
        return update_nested_dict(d_copy, u_copy)

def get_class_that_defined_method(meth):
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (inspect.isbuiltin(meth) and getattr(meth, '__self__', None) is not None and getattr(meth.__self__, '__class__', None)):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
                      None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)


from typing import Optional, Union, Dict, List, Tuple
from copy import deepcopy
import re
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.transforms as transforms
from matplotlib.patches import Rectangle
from matplotlib.lines import Line2D
from matplotlib.ticker import (MultipleLocator, FormatStrFormatter,
                               AutoMinorLocator, ScalarFormatter,
                               Locator, Formatter, AutoLocator)

from quickstats.utils.common_utils import update_nested_dict

class NumericFormatter(ScalarFormatter):
    def __call__(self, x, pos=None):
        tmp_format = self.format
        if x.is_integer():
            self.format = re.sub(r"1\.\d+f", r"1.0f", self.format)
        result = super().__call__(x, pos)
        self.format = tmp_format
        return result

COLOR_PALLETE = ['#000000', '#F2385A', '#4AD9D9', '#FDC536', '#125125', '#E88EED', '#B68D40']


PLOT_STYLES = {
    'distributions': {
        'marker': 'o',
        'markersize': 1,
        'linewidth': 1
    }
}

TEMPLATE_STYLES = {
    'default': {
        'figure':{
            'figsize': (11.111, 8.333),
            'dpi': 72,
            'facecolor': "#FFFFFF"
        },
        'legend_Line2D': {
            'linewidth': 3
        },
        'annotation':{
            'fontsize': 12
        },
        'axis': {
            'major_length': 16,
            'minor_length': 8,
            'major_width': 2,
            'minor_width': 1,
            'spine_width': 2,
            'labelsize': 20,
            'tick_bothsides': True
        },
        'xtick':{
            'format': 'numeric',
            'steps': None,
            'prune': None
        },
        'ytick':{
            'format': 'numeric', 
            'steps': None,
            'prune': None
        },        
        'xlabel': {
            'fontsize': 22,
            'loc' : 'right',
            'labelpad': 10
        },
        'ylabel': {
            'fontsize': 22,
            'loc' : 'top',
            'labelpad': 15
        },
        'text':{
            'fontsize': 20,
        },
        'plot':{
            'linewidth': 2
        },
        'errorbar': {
            "marker": 'x',
            "linewidth": 0,
            "markersize": 0,
            "elinewidth": 1,
            "capsize": 2,
            "capthick": 1
        },
        'legend':{
            "fontsize": 20
        }
    },
    'limit': {
        'axis':{
            'tick_bothsides': False
        },
        'errorbar': {
            "linewidth": 1,
            "markersize": 5,
            "marker": 'o',
        }
    },
    'limit_point':{
        'figure':{
            'figsize': (11.111, 10.333),
            'dpi': 72,
            'facecolor': "#FFFFFF"
        },        
        'axis':{
            'tick_bothsides': False
        },
        'legend':{
            'fontsize': 22
        },
        'text':{
            'fontsize': 22
        }
    },
    'teststat_dist': {
        'legend': {
            'fontsize': 12
        },
        'annotation':{
            'fontsize': 12
        }
    }
}

ANALYSIS_OPTIONS = {
    'default': {
        'loc': (0.05, 0.95),
        'fontsize': 20
    },
    'Run2': {
        'status': 'int', 
        'energy' : '13 TeV', 
        'lumi' : 139,
    }
}

def parse_styles(styles:Optional[Union[Dict, str]]=None):
    default_styles = deepcopy(TEMPLATE_STYLES['default'])
    if styles is None:
        styles = default_styles
    elif isinstance(styles, str):
        template_styles = TEMPLATE_STYLES.get(styles, None)
        if template_styles is None:
            raise ValueError(f"template styles `{styles}` not found")
        styles = update_nested_dict(default_styles, deepcopy(template_styles))
    else:
        styles = update_nested_dict(default_styles, deepcopy(styles))
    return styles

def parse_analysis_label_options(options:Optional[Dict]=None):
    default_options = deepcopy(ANALYSIS_OPTIONS['default'])
    if options is None:
        options = default_styles
    elif isinstance(options, str):
        template_options = ANALYSIS_OPTIONS.get(options, None)
        if template_options is None:
            raise ValueError(f"template analysis label options `{styles}` not found")
        options = update_nested_dict(default_options, deepcopy(template_options))
    else:
        options = update_nested_dict(default_options, deepcopy(options))
    return options

def single_frame(logx:bool=False, logy:bool=False, 
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Union[Dict, str]]=None):
    plt.clf()
    styles = parse_styles(styles)
    fig, ax = plt.subplots(nrows=1, ncols=1, **styles['figure'])
    if logx:
        ax.set_xscale('log')
    if logy:
        ax.set_yscale('log')
    format_axis_ticks(ax, x_axis=True, y_axis=True, xtick_styles=styles['xtick'],
                      ytick_styles=styles['ytick'], **styles['axis'])
        
    if analysis_label_options is not None:
        draw_atlas_label(ax, dy=0.05, text_options=styles['text'], **analysis_label_options)
        
    return ax


def suggest_markersize(nbins:int):
    if nbins <= 20:
        return 10
    elif (nbins > 20) and (nbins <= 200):
        return nbins/20
    return 1

def format_axis_ticks(ax, x_axis=True, y_axis=True, major_length:int=16, minor_length:int=8,
                      spine_width:int=2, major_width:int=2, minor_width:int=1, 
                      label_bothsides:bool=False, tick_bothsides:bool=False,
                      labelsize:Optional[int]=None,
                      x_axis_styles:Optional[Dict]=None, 
                      y_axis_styles:Optional[Dict]=None,
                      xtick_styles:Optional[Dict]=None,
                      ytick_styles:Optional[Dict]=None):
    if x_axis:
        if (ax.get_xaxis().get_scale() != 'log'):
            ax.xaxis.set_minor_locator(AutoMinorLocator())
        styles = {"labelsize":labelsize}
        styles['labeltop'] = label_bothsides
        styles['labelbottom'] = True
        styles['top'] = tick_bothsides
        styles['bottom'] = True
        if x_axis_styles is not None:
            styles.update(x_axis_styles)
        ax.tick_params(axis="x", which="major", direction='in', length=major_length,
                       width=major_width, **styles)
        ax.tick_params(axis="x", which="minor", direction='in', length=minor_length,
                       width=minor_width, **styles)
    if y_axis:
        if (ax.get_yaxis().get_scale() != 'log'):
            ax.yaxis.set_minor_locator(AutoMinorLocator())    
        styles = {"labelsize":labelsize}
        styles['labelleft'] = True
        styles['labelright'] = label_bothsides
        styles['left'] = True
        styles['right'] = tick_bothsides
        if y_axis_styles is not None:
            styles.update(y_axis_styles)
        ax.tick_params(axis="y", which="major", direction='in', length=major_length,
                       width=major_width, **styles)
        ax.tick_params(axis="y", which="minor", direction='in', length=minor_length,
                       width=minor_width, **styles)
    for axis in ['top','bottom','left','right']:
        ax.spines[axis].set_linewidth(spine_width)
        
    set_axis_tick_styles(ax.xaxis, xtick_styles)
    set_axis_tick_styles(ax.yaxis, ytick_styles)

def set_axis_tick_styles(ax, styles=None):
    if styles is None:
        return None
    fmt = styles['format']
    if fmt is not None:
        formatter = None
        if isinstance(fmt, str):
            if fmt == 'numeric':
                formatter = NumericFormatter()
        if isinstance(fmt, Formatter):
            formatter = fmt
        if formatter is None:
            raise ValueError(f"unsupported axis tick format {fmt}")
        ax.set_major_formatter(formatter)
    locator = AutoLocator()
    locator_params = {}
    for param in ['steps', 'prune']:
        value = styles[param]
        if value is not None:
            locator_params[param] = value
    locator.set_params(**locator_params)
    ax.set_major_locator(locator)
                

def centralize_axis(ax, which='x'):
    if which == 'x':
        limits = ax.get_xlim()
        xmax = np.max(np.abs(limits))
        ax.set_xlim(-xmax, xmax)
    elif which == 'y':
        limits = ax.get_ylim()
        ymax = np.max(np.abs(limits))
        ax.set_ylim(-ymax, ymax)
        
def parse_transform(obj:str):
    if not obj:
        transform = None
    elif obj == 'figure':
        transform = plt.gcf().transFigure
    elif obj == 'axis':
        transform = plt.gca().transAxes
    elif obj == 'data':
        transform = plt.gca().transData
    return transform

def create_transform(transform_x:str='axis', transform_y:str='axis'):
    transform = transforms.blended_transform_factory(parse_transform(transform_x), 
                                                     parse_transform(transform_y))
    return transform

def get_box_dimension(box):
    axis = plt.gca()
    plt.gcf().canvas.draw()
    bb = box.get_window_extent()
    points  = bb.transformed(axis.transAxes.inverted()).get_points().transpose()
    xmin = np.min(points[0])
    xmax = np.max(points[0])
    ymin = np.min(points[1])
    ymax = np.max(points[1])
    return xmin, xmax, ymin, ymax

def draw_sigma_bands(axis, ymax, height=1.0):
    # +- 2 sigma band
    axis.add_patch(Rectangle((-2, -height/2), 2*2, ymax + height/2, fill=True, color='yellow'))
    # +- 1 sigma band
    axis.add_patch(Rectangle((-1, -height/2), 1*2, ymax + height/2, fill=True, color='lime'))
    
def draw_sigma_lines(axis, ymax, height=1.0, **styles):
    y = [-height/2, ymax*height - height/2]
    axis.add_line(Line2D([-1, -1], y, **styles))
    axis.add_line(Line2D([+1, +1], y, **styles))
    axis.add_line(Line2D([0, 0], y, **styles)) 
    
def draw_hatches(axis, ymax, height=1.0, **styles):
    x_min    = axis.get_xlim()[0]
    x_max    = axis.get_xlim()[1]
    x_range  = x_max - x_min
    y_values = np.arange(0, height*ymax, 2*height) - height/2
    for y in y_values:
        axis.add_patch(Rectangle((x_min, y), x_range, 1, **styles, zorder=-1))

def draw_text(axis, x, y, s, transform_x:str='axis', 
              transform_y:str='axis', **styles):
    transform = transforms.blended_transform_factory(parse_transform(transform_x), 
                                                     parse_transform(transform_y))
    text = axis.text(x, y, s, transform=transform, **styles)
    xmin, xmax, ymin, ymax = get_box_dimension(text)
    return xmin, xmax, ymin, ymax


def draw_ATLAS_label(axis, loc=(0.05, 0.95), fontsize=25, extra='Internal', 
                     transform_x='axis', transform_y='axis', 
                     vertical_align='top', horizontal_align='left'):
    current_axis = plt.gca()
    plt.sca(axis)
    if vertical_align not in ['top', 'bottom']:
        raise ValueError('only "top" or "bottom" vertical alignment is allowed')
    if horizontal_align not in ['left', 'right']:
        raise ValueError('only "left" or "right" horizontal alignment is allowed') 
    transform = transforms.blended_transform_factory(parse_transform(transform_x), 
                                                     parse_transform(transform_y))
    x, y = loc
    text_atlas = axis.text(x, y, 'ATLAS', fontsize=fontsize, transform=transform,
                           horizontalalignment=horizontal_align,
                           verticalalignment=vertical_align,
                           fontproperties={"weight":"bold", "style":"italic"})
    xmin, xmax, ymin, ymax = get_box_dimension(text_atlas)
    text_width = xmax - xmin
    dx = text_width/15
    text_extra = axis.text(xmax + dx, ymin, extra, fontsize=fontsize, transform=axis.transAxes,
                           horizontalalignment='left', verticalalignment='bottom')
    plt.sca(current_axis)
    _, xmax, _, ymax = get_box_dimension(text_atlas)
    return xmin, xmax, ymin, ymax

def draw_atlas_label(axis, loc=(0.05, 0.95), fontsize=25, status:str='int',
                     energy:Optional[str]=None, lumi:Optional[str]=None,
                     extra_text:Optional[str]=None, dy:float=0.05,
                     transform_x:str='axis', transform_y:str='axis',
                     text_options:Optional[Dict]=None):
    
    if status == "final":
        status_str = ""
    elif status == "int":
        status_str = "Internal"
    elif status == "wip":
        status_str = "Work in Progress"
    elif status == "prelim":
        status_str = "Preliminary"
    elif status == "opendata":
        status_str = "Open Data"
    else:
        status_str = status
        
    xmin, xmax, ymin, ymax = draw_ATLAS_label(axis, loc, fontsize, extra=" "+status_str,
                                              transform_x=transform_x,
                                              transform_y=transform_y)

    elumi_text = []
    if energy is not None:
        elumi_text.append(r"$\sqrt{s} = $ " + energy )
    if lumi is not None:
        elumi_text.append(lumi)
    elumi_text = ", ".join(elumi_text)
    
    texts = []
    if elumi_text:
        texts.append(elumi_text)
        
    if extra_text is not None:
        texts += extra_text.split("//")
    
    if text_options is None:
        text_options = {}
    
    for text in texts:
        _, _, ymin, _ = draw_text(axis, xmin, ymin - dy, text, transform_x=transform_x,
                                  transform_y=transform_y, **text_options)
        
    return

def add_collective_data(ax, data, plot_styles=None):
    if plot_styles is None:
        plot_styles = None
    color_iter = iter(COLOR_PALLETE)
    for label in data:
        styles = plot_styles.copy()
        x = data[label]['x']
        y = data[label]['y']
        nbins = len(x)
        if 'markersize' not in styles:
            styles['markersize'] = suggest_markersize(nbins)        
        if 'color' not in styles:
            styles['color'] = next(color_iter)
        styles['label'] = label
        ax.plot(x, y, **styles)

def plot_distributions(data, xlabel='', ylabel='', 
                       figsize=(14, 8), plot_styles=None, atlas_label=False):
    plt.clf()
    base_size = figsize[0]
    fig = plt.figure(figsize=figsize)
    ax = plt.gca()
    styles = PLOT_STYLES['distributions'].copy()
    if plot_styles is not None:
        styles.update(plot_styles)
    add_collective_data(ax, data, styles)
    limits = ax.get_ylim()
    ax.set_ylim(limits[0], limits[1]*1.2)
    ax.set_xlabel(xlabel, fontsize=base_size*1.6,  labelpad=base_size*0.7)
    ax.set_ylabel(ylabel, fontsize=base_size*1.6,  labelpad=base_size*0.7)
    format_axis_ticks(ax, major_length=14, minor_length=7, major_width=1.3, 
                      minor_width=1.3, labelsize=base_size*1.4)
    if atlas_label:
        draw_ATLAS_label(ax, fontsize=base_size*1.8)
    ax.legend(fontsize=base_size*1.2)
    return plt


import quickstats

from .color_schemes import *
from .abstract_plot import AbstractPlot
from .multi_data_plot import MultiDataPlot
from .hypotest_inverter_plot import HypoTestInverterPlot
from .score_distribution_plot import score_distribution_plot
from .test_statistic_distribution_plot import TestStatisticDistributionPlot
from .upper_limit_1D_plot import UpperLimit1DPlot
from .upper_limit_2D_plot import UpperLimit2DPlot
from .likelihood_1D_plot import Likelihood1DPlot

from matplotlib import style, colors

# Reference from https://github.com/beojan/atlas-mpl

style.core.USER_LIBRARY_PATHS.append(quickstats.stylesheet_path)
style.core.reload_library()
style.use("quick_default")

colors.EXTRA_COLORS = EXTRA_COLORS
colors.colorConverter.colors.update(EXTRA_COLORS)


from typing import Optional, Union, Dict, List

import matplotlib.patches as patches
import matplotlib.lines as lines
import numpy as np
import pandas as pd

from quickstats.plots.template import single_frame, parse_styles, format_axis_ticks, create_transform
from quickstats.plots import AbstractPlot
from quickstats.utils.common_utils import combine_dict

class UpperLimit1DPlot(AbstractPlot):
    
    COLOR_PALLETE = {
        '2sigma': 'hh:darkyellow',
        '1sigma': 'hh:lightturquoise',
        'expected': 'k',
        'observed': 'k',
    }
    
    LABELS = {
        '2sigma': r'Expected limit $\pm 2\sigma$',
        '1sigma': r'Expected limit $\pm 1\sigma$',
        'expected': r'Expected limit',
        'observed': r'Observed limit',
    }    
    
    def __init__(self, category_df, label_map, line_below=None,
                 color_pallete:Optional[Dict]=None,
                 labels:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]='limit_point',
                 analysis_label_options:Optional[Union[Dict, str]]=None):
        super().__init__(color_pallete=color_pallete,
                         styles=styles,
                         analysis_label_options=analysis_label_options)
        self.category_df = category_df
        self.label_map = label_map
        self.line_below = line_below
        
        self.labels = combine_dict(self.LABELS, labels)  
    
    def draw(self, logx:bool=False, xlabel:Optional[str]=None, markersize:float=50.,
             draw_observed:bool=True, draw_stat:bool=False, sig_fig:int=2):
        if (draw_observed + draw_stat) > 1:
            raise RuntimeError("draw_observed and draw_stat can not be both True")
        n_category = len(self.category_df.columns)
        ax = self.draw_frame(logx=logx)
        transform = create_transform(transform_x='axis', transform_y='data')
        
        if draw_observed:
            text_pos = {'observed': 0.775, 'expected': 0.925}
        if draw_stat:
            text_pos = {'expected': 0.775, 'stat': 0.925}
        if (not draw_observed) and (not draw_stat):
            text_pos = {'expected': 0.925}
        
        for i, category in enumerate(self.category_df):
            df = self.category_df[category]
            # draw observed
            if draw_observed:
                observed_limit = df['obs']
                ax.vlines(observed_limit, i, i+1, colors=self.color_pallete['observed'], linestyles='solid', 
                          zorder=1.1, label=self.labels['observed'] if i == 0 else '')
                ax.scatter(observed_limit, i + 0.5, s=markersize, marker='o', 
                           color=self.color_pallete['observed'], zorder=1.1)
                ax.text(text_pos['observed'], i + 0.5, f"{{:.{sig_fig}f}}".format(observed_limit), 
                        horizontalalignment='center', 
                        verticalalignment='center',
                        transform=transform,
                        **self.styles['text'])
            # draw stat
            if draw_stat:
                stat_limit = df['stat']
                ax.text(text_pos['stat'], i + 0.5, f"({{:.{sig_fig}f}})".format(stat_limit), 
                        horizontalalignment='center', 
                        verticalalignment='center',
                        transform=transform,
                        **self.styles['text'])
            # draw expected
            expected_limit = df['0']
            ax.vlines(expected_limit, i, i + 1, colors=self.color_pallete['expected'], linestyles = 'dotted',
                      zorder = 1.1, label=self.labels['expected'] if i==0 else '')
            ax.text(text_pos['expected'], i + 0.5, f"{{:.{sig_fig}f}}".format(expected_limit), 
                    horizontalalignment='center', 
                    verticalalignment='center',
                    transform=transform,
                    **self.styles['text'])
            # draw +1
            ax.fill_betweenx([i, i + 1], df['-2'], df['2'], facecolor=self.color_pallete['2sigma'], 
                             label=self.labels['2sigma'] if i==0 else '')
            ax.fill_betweenx([i, i + 1], df['-1'], df['1'], facecolor=self.color_pallete['1sigma'], 
                             label=self.labels['1sigma'] if i==0 else '')
        xlim = ax.get_xlim()
        ax.set_xlim(xlim[0] - (xlim[1]/0.7 - xlim[1])*0.5, xlim[1]/0.7)
        ax.set_ylim(0, len(self.category_df.columns) + 2.2)
        ax.set_yticks(np.arange(n_category) + 0.5, minor=False)
        ax.tick_params(axis="y", which="minor", length=0)
        for axis in ['top', 'bottom', 'left', 'right']:
            ax.spines[axis].set_linewidth(2)
        ax.set_yticklabels([self.label_map[i] for i in self.category_df.columns.to_list()], 
                           horizontalalignment='right')
        # draw horizonal dashed lines
        ax.axhline(n_category, color = 'k', ls = '--', lw=1)
        if self.line_below is not None:
            for category in self.line_below:
                position = np.where(np.array(self.category_df.columns, dtype='str') == category)[0]
                if position.shape[0] != 1:
                    raise ValueError("category `{}` not found in dataframe".format(category))
                ax.axhline(position[0], color = 'k', ls = '--', lw=1)
        if draw_observed:
            ax.text(text_pos['observed'], n_category + 0.3, 'Obs.', 
                    horizontalalignment='center', 
                    verticalalignment='center',
                    transform=transform,
                    **self.styles['text'])
        if draw_stat:
            ax.text(text_pos['stat'], n_category + 0.3, '(Stat.)', 
                    horizontalalignment='center', 
                    verticalalignment='center',
                    transform=transform,
                    **self.styles['text'])
        ax.text(text_pos['expected'], n_category + 0.3, 'Exp.', 
                horizontalalignment='center', 
                verticalalignment='center',
                transform=transform,
                **self.styles['text'])
        if xlabel is not None:
            ax.set_xlabel(xlabel, **self.styles['xlabel'])
        # border for the legend
        border_leg = patches.Rectangle((0, 0), 1, 1, facecolor = 'none', edgecolor = 'black', linewidth = 1)

        handles, labels = ax.get_legend_handles_labels()
        if draw_observed:
            handles = [handles[0], handles[1], (handles[3], border_leg), (handles[2], border_leg)]
            labels  = [labels[0], labels[1], labels[3], labels[2]]
        if draw_stat:
            handles = [handles[0], (handles[2], border_leg), (handles[1], border_leg)]
            labels  = [labels[0], labels[2], labels[1]]
        ax.legend(handles, labels, **self.styles['legend'])
        return ax


from typing import Optional, List, Dict, Union
import os
import glob
import json
import time
import ROOT

from .builtin_methods import BUILTIN_METHODS
from .actions import *
from .parsers import RooProcConfigParser

from quickstats.components import AbstractObject

class RooProcessor(AbstractObject):
    def __init__(self, config_path:Optional[str]=None,
                 multithread:bool=True,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)
        self.action_list = []
        self.rdf_frames = {}
        self.rdf = None
        self.global_variables = {}
        self.external_variables = {}
        self.treename = None
        
        self.load_buildin_functions()
        
        if multithread:
            ROOT.EnableImplicitMT()
        
        if config_path is not None:
            self.load_config(config_path)
            
    def load_buildin_functions(self):
        for name, definition in BUILTIN_METHODS.items():
            RooProcDeclare.declare_expression(definition, name)
    
    def load_config(self, config_path:Optional[str]=None):
        action_list = RooProcConfigParser.parse_file(config_path)
        if len(action_list) == 0:
            raise RuntimeError("no actions found in the process card")
        first_action = action_list[0]
        if not isinstance(first_action, RooProcTreeName):
            raise RuntimeError("tree name must be specified at the beginning of the process card")
        self.treename = first_action._params['treename']
        self.action_list = action_list
    
    def run(self, filename:str):
        if os.path.isdir(filename):
            all_files = glob.glob(os.path.join(filename, "*.root"))
        else:
            all_files = glob.glob(filename)
        if not all_files:
            raise FileNotFoundError(f"file `{filename}` does not exist")
        if len(all_files) == 1:
            self.stdout.info(f"INFO: Processing file `{all_files[0]}`.")
        else:
            self.stdout.info(f"INFO: Professing files")
            for f in all_files:
                self.stdout.info(f"\t`{f}`")
        if len(self.action_list) == 0:
            self.stdout.warning("WARNING: No actions to be performed.")
            return None
        if self.treename is None:
            raise RuntimeError("tree name is undefined")
        start = time.time()
        self.rdf = ROOT.RDataFrame(self.treename, all_files)
        for i, action in enumerate(self.action_list):
            if isinstance(action, RooProcRDFAction):
                self.rdf = action.execute(self.rdf, self.global_variables)
            elif isinstance(action, RooProcHelperAction):
                action.execute(self, self.global_variables)
            elif isinstance(action, RooProcHybridAction):
                self.rdf, _ = action.execute(self.rdf, self, self.global_variables)
            else:
                raise RuntimeError("unknown action type")
        end = time.time()
        time_taken = end - start
        self.stdout.info(f"INFO: Task finished. Total time taken: {time_taken:.3f} s.")


from typing import Optional, Union, Dict, List, Tuple
import math

import numpy as np

import ROOT

from quickstats.plots import AbstractPlot
from quickstats.plots.template import single_frame, parse_styles
from quickstats.utils.roostats_utils import get_hypotest_data

class TestStatisticDistributionPlot(AbstractPlot):
    
    COLOR_PALLETE = {
        'null': 'b',
        'alt': 'r',
        'null_sec': 'tab:cyan',
        'alt_sec': 'tab:pink'
    }
        
    def __init__(self, result:ROOT.RooStats.HypoTestResult,
                 sec_result:Optional[ROOT.RooStats.HypoTestResult]=None,
                 teststat_str:str=r"\~{q}_{\mu}",
                 sec_teststat_str:str=r"\~{q}_{\mu}",
                 annotate_str:str=r"toys",
                 sec_annotate_str:str=r"toys",
                 color_pallete:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]='teststat_dist',
                 analysis_label_options:Optional[Union[Dict, str]]='default'):
        super().__init__(color_pallete=color_pallete, styles=styles,
                         analysis_label_options=analysis_label_options)
        
        self.result = result
        self.data = get_hypotest_data(self.result)
        self.teststat_str = teststat_str
        self.annotate_str = annotate_str
        
        # secondary result (for comparison plot)
        self.sec_result = sec_result
        if self.sec_result is None:
            self.sec_data = None
        else:
            self.sec_data = get_hypotest_data(self.sec_result)
        self.sec_teststat_str = sec_teststat_str
        self.sec_annotate_str = sec_annotate_str
        
        self.asymptotic_mu = None
        self.asymptotic_sigma = None
        
    def draw_distribution(self, ax, data, weight, label:str,
                          color:Optional[str]=None,
                          nbins:int=75, xmax:float=40.):
        n, bins   = np.histogram(data, bins=nbins, range=(0, xmax), 
                                 density=False, weights=weight/weight.sum())
        bin_centers  = 0.5*(bins[1:] + bins[:-1])
        bin_width    = bin_centers[1] - bin_centers[0]
        h = ax.errorbar(bin_centers, n, color=color, xerr=0.5*bin_width, 
                        yerr = (n/weight.sum())**0.5, **self.styles['errorbar'], 
                        label=label, zorder=999)
        return h
    
    @staticmethod
    def get_qmutilde_cdf(qmutilde, mu:float, muprime:float, sigma:float, scale_factor:float=1.):
        mu2 = mu * mu
        sigma2 = sigma * sigma
        mu2oversigma2 = mu2 / sigma2
        if (muprime != mu):
            if (qmutilde < mu2oversigma2):
                value = 0.5 * (1. / np.sqrt(2. * np.pi)) * (1. / np.sqrt(qmutilde)) * \
                        np.exp(-0.5 * np.power(np.sqrt(qmutilde) - (mu - muprime) / sigma, 2))
            else:
                value = 0.5 * sigma * (1. / mu) * (1. / np.sqrt(2. * np.pi)) * \
                        np.exp(-0.5 * np.power((qmutilde - (mu2 - 2 * mu * muprime) / (sigma2)) / (2. * mu / sigma), 2))
        else:
            if (qmutilde < mu2oversigma2):
                value = 0.5 * (1. / np.sqrt(2. * np.pi)) * (1. / np.sqrt(qmutilde)) * \
                        np.exp(-0.5 * qmutilde)
            else:
                value = 0.5 * sigma * (1. / mu) * (1. / np.sqrt(2. * np.pi)) * \
                        np.exp(-0.5 * np.power((qmutilde + mu2oversigma2) / (2. * mu / sigma), 2))
        value *= scale_factor
        return value
    
    def set_asymptotic_params(self, mu:Optional[float]=None, sigma:Optional[float]=None):
        self.asymptotic_mu    = mu
        self.asymptotic_sigma = sigma
    
    def draw_asymptotic(self, ax, xmax:float, nbins:int):
        if (self.asymptotic_mu is None) or (self.asymptotic_sigma is None):
            raise RuntimeError(" mu and sigma values for asymptotic results are not initialized")
        bins = np.linspace(0, xmax, nbins+1)
        bin_centers  = 0.5*(bins[1:] + bins[:-1])
        bin_width    = bin_centers[1] - bin_centers[0]
        mu = self.asymptotic_mu
        sigma = self.asymptotic_sigma
        asym_curve_null = np.array([self.get_qmutilde_cdf(i, mu, mu, sigma, bin_width) for i in bin_centers])
        asym_curve_alt  = np.array([self.get_qmutilde_cdf(i, mu, 0, sigma, bin_width) for i in bin_centers])
        h_asym_null = ax.plot(bin_centers, asym_curve_null, 
                              color=self.color_pallete['null'],
                              label=fr"$f({self.teststat_str}|\mu)$ asymptotic")
        h_asym_alt  = ax.plot(bin_centers, asym_curve_alt, 
                              color=self.color_pallete['alt'],
                              label=fr"$f({self.teststat_str}|0)$ asymptotic")
        return h_asym_null, h_asym_alt

    def draw_annotation(self, ax, data, xmin:float=0.35, ymax:float=0.9):
        null_converged = data['null']['size']
        alt_converged = data['alt']['size']        
        obs_pmu = data['observed']['CLsplusb']
        obs_pmu_err = data['observed']['CLsplusbError']
        obs_pb = 1 - data['observed']['CLb']
        obs_pb_err = data['observed']['CLbError']
        obs_CLs = data['observed']['CLs']
        obs_CLs_err = data['observed']['CLsError']
        exp_pmu = data['expected']['CLsplusb']
        exp_pmu_err = data['expected']['CLsplusbError']
        exp_pb = 1 - data['expected']['CLb']
        exp_pb_err = data['expected']['CLbError']
        exp_CLs = data['expected']['CLs']
        exp_CLs_err = data['expected']['CLsError']
        ax.annotate('\n'.join([fr'     # converged null : {null_converged}',
                    fr'     # converged alt   : {alt_converged}',
                    fr'$p_{{\mu}}^{{obs}}$(toys)    = {obs_pmu:.4f} $\pm$ {obs_pmu_err:.4f}',
                    fr'$p_{{b}}^{{obs}}$(toys)    = {obs_pb:.4f} $\pm$ {obs_pb_err:.4f}',
                    fr'$CL_s^{{obs}}$(toys)  = {obs_CLs:.4f} $\pm$ {obs_CLs_err:.4f}',
                    fr'$p_{{\mu}}^{{exp}}$(toys)    = {exp_pmu:.4f} $\pm$ {exp_pmu_err:.4f}',
                    fr'$p_{{b}}^{{exp}}$(toys)    = {exp_pb:.4f} $\pm$ {exp_pb_err:.4f}',
                    fr'$CL_s^{{exp}}$(toys)  = {exp_CLs:.4f} $\pm$ {exp_CLs_err:.4f}']),
                    (xmin, ymax), xycoords='axes fraction', 
                    horizontalalignment='left', verticalalignment='top',
                    bbox={'facecolor': 'white', 'edgecolor': 'white', 'pad': 5, 'alpha':0.8},
                    **self.styles['annotation'])
        

    def draw_quantiles(self, ax, data, teststat_str:str=r"\~{q}_{\mu}", 
                       linestyle:str='-', ymax:float=0.7):
        qmu_obs = data['observed']['teststat']
        qmu_exp = data['expected']['teststat'][0]
        qmu_exp_n2s = data['expected']['teststat'][-2]
        qmu_exp_n1s = data['expected']['teststat'][-1]
        qmu_exp_p1s = data['expected']['teststat'][1]
        qmu_exp_p2s = data['expected']['teststat'][2]
        h_obs = ax.axvline(qmu_obs, 0, ymax, color='k', linestyle=linestyle,
                           label=fr"${teststat_str}^{{obs}}$")
        h_n2s = ax.axvline(qmu_exp_n2s, 0, ymax, color='#ffcc00', linestyle=linestyle,
                           label=fr"${teststat_str}^{{exp}}(-2\sigma)$", clip_on=False)
        h_n1s = ax.axvline(qmu_exp_n1s, 0, ymax, color='#33ad33', linestyle=linestyle,
                           label=fr"${teststat_str}^{{exp}}(-1\sigma)$")
        h_exp = ax.axvline(qmu_exp, 0, ymax, color='#52ffff', linestyle=linestyle,
                           label=fr"${teststat_str}^{{exp}}$")
        h_p1s = ax.axvline(qmu_exp_p1s, 0, ymax, color='#33ad33', linestyle=linestyle,
                           label=fr"${teststat_str}^{{exp}}(+1\sigma)$")
        h_p2s = ax.axvline(qmu_exp_p2s, 0, ymax, color='#ffcc00', linestyle=linestyle,
                           label=fr"${teststat_str}^{{exp}}(+2\sigma)$")
        # return the plot handles
        return [h_obs, h_n2s, h_n1s, h_exp, h_p1s, h_p2s]
    
    def draw_legends(self, ax, primary_handles, secondary_handles=None, asymptotic_handles=None,
                     leg_loc:Tuple=(1.02, 0.75), sec_leg_loc:Tuple=(1.02, 0.5)):
        handles, labels = ax.get_legend_handles_labels()
        if asymptotic_handles is not None:
            primary_handles += asymptotic_handles
        primary_labels = [labels[handles.index(h)] for h in primary_handles]
        primary_leg = ax.legend(primary_handles, primary_labels, 
                                loc=leg_loc, ncol=2, **self.styles['legend'])
        ax.add_artist(primary_leg)
        
        if secondary_handles is not None:
            secondary_labels = [labels[handles.index(h)] for h in secondary_handles]
            primary_leg = ax.legend(secondary_handles, secondary_labels, 
                                    loc=leg_loc, ncol=2, **self.styles['legend'])
            ax.add_artist(primary_leg)
        
        
    def draw(self, xlabel:str=r"$\~{q}_{\mu}$",
             nbins:int=75, xmax:Optional[float]=None, ymax:float=1e3,
             leg_loc:Tuple=(1.02, 0.7), sec_leg_loc:Tuple=(1.02, 0.45)):
        if self.sec_data is not None:
            annotation_ymax = 0.9
        else:
            annotation_ymax = 0.95
        data_max = max([np.max(self.data['null']['data']), np.max(self.data['alt']['data'])])
        if self.sec_data is not None:
            data_max_sec = max(np.max(self.sec_data['null']['data']), np.max(self.sec_data['alt']['data']))
            data_max = max([data_max, data_max_sec])
        if xmax is None:
            xmax = math.ceil(data_max * 1.1)
        binwidth = xmax / nbins
        ax = self.draw_frame(logy=True)
        ax.set_xlabel(xlabel, **self.styles['xlabel'])
        ax.set_ylabel("Normalized Entries / {:.3f}".format(binwidth), **self.styles['ylabel'])
        
        # draw test statistic distribution for null hypothesis
        h_null = self.draw_distribution(ax, 
                                       self.data['null']['data'],
                                       self.data['null']['weight'],
                                       label=fr"$f({self.teststat_str}|\mu)$ {self.annotate_str}",
                                       color=self.color_pallete['null'],
                                       nbins=nbins, xmax=xmax)
        # draw test statistic distribution for alternative hypothesis
        h_alt = self.draw_distribution(ax, 
                               self.data['alt']['data'],
                               self.data['alt']['weight'],
                               label=fr"$f({self.teststat_str}|0)$ {self.annotate_str}",
                               color=self.color_pallete['alt'],
                               nbins=nbins, xmax=xmax)
        h_quantiles = self.draw_quantiles(ax, self.data, teststat_str=self.teststat_str)
        self.draw_annotation(ax, self.data, xmin=0.4, ymax=annotation_ymax)
        primary_handles = h_quantiles + [h_null, h_alt]
        
        if self.sec_data is not None:
            h_null_sec = self.draw_distribution(ax, 
                                           self.sec_data['null']['data'],
                                           self.sec_data['null']['weight'],
                                           label=fr"$f({self.sec_teststat_str}|\mu)$ {self.sec_annotate_str}",
                                           color=self.color_pallete['null_sec'],
                                           nbins=nbins, xmax=xmax)
            h_alt_sec = self.draw_distribution(ax, 
                                   self.sec_data['alt']['data'],
                                   self.sec_data['alt']['weight'],
                                   label=fr"$f({self.sec_teststat_str}|0)$ {self.sec_annotate_str}",
                                   color=self.color_pallete['alt_sec'],
                                   nbins=nbins, xmax=xmax)
            h_quantiles_sec = self.draw_quantiles(ax, self.sec_data, 
                                                  teststat_str=self.sec_teststat_str,
                                                  linestyle='--')
            self.draw_annotation(ax, self.sec_data, xmin=0.7, ymax=annotation_ymax)
            secondary_handles = h_quantiles_sec + [h_null_sec, h_alt_sec]
            
            ax.annotate(f'${self.teststat_str}$ {self.annotate_str}',
                        (0.45, 0.95), xycoords='axes fraction', verticalalignment='top', 
                        horizontalalignment='left', **self.styles['annotation'])
            ax.annotate(f'${self.sec_teststat_str}$ {self.sec_annotate_str}',
                        (0.75, 0.95), xycoords = 'axes fraction', verticalalignment='top', 
                        horizontalalignment='left', **self.styles['annotation'])
        else:
            secondary_handles = None

        ymin = ax.get_ylim()[0]
        
        # draw test statistic distribution from asymptotic approximation
        if (self.asymptotic_mu is not None) and (self.asymptotic_sigma is not None):
            h_asym_null, h_asym_alt = self.draw_asymptotic(ax, xmax, nbins)
            asymptotic_handles = [h_asym_null[0], h_asym_alt[0]]
        else:
            asymptotic_handles = None
            
        ax.set_xlim([0, xmax])
        ax.set_ylim([ymin, ymax])
        
        self.draw_legends(ax, primary_handles, secondary_handles, asymptotic_handles,
                          leg_loc=leg_loc, sec_leg_loc=sec_leg_loc)
        
        return ax


from typing import Optional, Union, Dict
import os
import math
import json
from itertools import repeat

import numpy as np

import quickstats
from quickstats.components import Likelihood
from quickstats.utils.common_utils import parse_config, execute_multi_tasks, NpEncoder

import ROOT

class PValueToys():
    
    def __init__(self, filename:str, poi_name:Optional[str]=None,
                 poi_val:Optional[float]=0, binned:bool=True,
                 verbosity:Union[int, str]="ERROR",
                 fit_options:Optional[Union[Dict, str]]=None):
        self.filename = filename
        self.poi_name = poi_name
        self.poi_val = poi_val
        self.binned = binned
        self.fit_options = parse_config(fit_options)
        self.fit_options['verbosity'] = verbosity
        
    def get_properties(self, nll_mufix:float, nll_muhat:float):
        q_mu = 2*(nll_mufix - nll_muhat)
        ndof = 1
        #p_value = ROOT.Math.chisquared_cdf_c(q_mu, ndof)
        pvalue = 1 - ROOT.Math.normal_cdf(math.sqrt(q_mu), 1, 0)
        significance = ROOT.RooStats.PValueToSignificance(pvalue)
        properties = {
            'q_mu': q_mu,
            'pvalue': pvalue,
            'significance': significance
        }
        return properties
    
    def _get_toy_results(self, n_toys:int, seed:int=0, batch_index:Optional[int]=None,
                         save_as:Optional[str]=None):
        if (batch_index is not None) and (save_as is not None):
            print(f"INFO: Evaluating toys for the batch: {batch_index}")
        likelihood = Likelihood(filename=self.filename,
                                poi_name=self.poi_name,
                                data_name=None,
                                config=self.fit_options)
        
        toys = likelihood.model.generate_toys(n_toys=n_toys,
                                              seed=seed,
                                              binned=self.binned,
                                              do_import=False)
        results = []
        for i in range(n_toys):
            likelihood.minimizer.data = toys[i]
            nll_muhat = likelihood.evaluate(poi_val=self.poi_val, unconditional=True)
            muhat = likelihood.poi.getVal()
            fit_status_muhat = likelihood.minimizer.status
            nll_mufix = likelihood.evaluate(poi_val=self.poi_val, unconditional=False)
            fit_status_mufix = likelihood.minimizer.status
            properties = self.get_properties(nll_mufix, nll_muhat)
            result = {
                'toy_index': i,
                'muhat': muhat, 
                'mufix': self.poi_val,
                'nll_muhat': nll_muhat,
                'nll_mufix': nll_mufix,
                'fit_status_muhat': fit_status_muhat,
                'fit_status_mufix': fit_status_mufix}
            result.update(properties)
            results.append(result)
        if batch_index is not None:
            for result in results:
                result['batch_index'] = batch_index
        if save_as is not None:
            with open(save_as, 'w') as f:
                json.dump(results, f, indent=2, cls=NpEncoder)
        return results
        
        
    def get_toy_results(self, n_toys:int=1000, seed:int=0, batchsize:int=100,
                        save_as:Optional[str]="toy_results.json", cache:bool=True,
                        parallel:int=-1):
        n_batches = n_toys // batchsize
        remainder = n_toys % batchsize
        if remainder != 0:
            batches = np.array([batchsize] * n_batches + [remainder])
        else:
            batches = np.array([batchsize] * n_batches)
        batch_indices = np.array(range(len(batches)))
        
        results = []
        
        if save_as:
            base_dir = os.path.dirname(save_as)
            if base_dir and (not os.path.exists(base_dir)):
                os.makedirs(base_dir, exist_ok=True)
            base_savename = os.path.splitext(save_as)[0]
            extension = os.path.splitext(save_as)[1]
            batch_save_as = [f"{base_savename}_batch_{i}{extension}" for i in batch_indices]
            cached_indices = []
            if cache:
                for i, path in enumerate(batch_save_as):
                    if os.path.exists(path):
                        try:
                            result = json.load(open(path, 'r'))
                            print(f'INFO: Cached batch toy results from "{path}"')
                            results += result
                            cached_indices.append(i)
                        except:
                            pass

            batches       = np.delete(batches, cached_indices)
            batch_indices = np.delete(batch_indices, cached_indices)
        else:
            batch_save_as = [None]*len(batches)
            
        arguments = (batches, repeat(seed), batch_indices, batch_save_as)
        
        
        tmp_verbosity = quickstats._PRINT_.verbosity

        batch_results = execute_multi_tasks(self._get_toy_results, *arguments, parallel=parallel)
        
        quickstats.set_verbosity(tmp_verbosity)
        
        for batch_result in batch_results:
            results += batch_result
        
        if save_as is not None:
            with open(save_as, 'w') as f:
                json.dump(results, f, indent=2, cls=NpEncoder)
        return results


import os
import time
from six import string_types

import numpy as np

import ROOT
import quickstats

def is_corrupt(f:ROOT.TFile):
    if f.IsZombie():
        return True
    if f.TestBit(ROOT.TFile.kRecovered):
        return True
    if f.GetNkeys() == 0:
        return True
    return False

def release_tfiles():
    for f in ROOT.gROOT.GetListOfFiles():
        f.Close()

def compile_macro(name:str):
    macros_dir = os.path.join(quickstats.macro_path, 'macros')
    macros_path = os.path.join(macros_dir, name + '.cxx')
    print('INFO: Compiling macro "{}"...'.format(name))
    if not os.path.exists(macros_path):
        print('ERROR: Cannot locate macro files from {}. ROOT macros will not be compiled.'.format(macros_dir))
        return -1
    return ROOT.gROOT.LoadMacro('{}++'.format(macros_path))

def load_macro(name:str):
    auto_load = ROOT.gInterpreter.AutoLoad(name)
    # already loaded by quickstats / third-party
    if auto_load == 1:
        return None
    macros_dir = os.path.join(quickstats.macro_path, 'macros')
    macros_path = os.path.join(macros_dir, name + '_cxx')
    if not os.path.exists(macros_path + '.so'):
        return compile_macro(name)
    result = ROOT.gSystem.Load(macros_path)
    if (result != 0) and (result != 1):
        raise RuntimeError("Shared library for the macro {} is incompatible with the current pyROOT version. "
                           "Please recompile by typing \"quickstats compile\".".format(name))
    return result


def print_resource_used(start):
    proc_info = ROOT.ProcInfo_t()
    flag = ROOT.gSystem.GetProcInfo(proc_info)
    if flag < 0:
        return None
    cpu_time = proc_info.fCpuUser + proc_info.fCpuSys
    wall_time = time.time()-start
    print('Resources Used: cpu_time={:.3f}s, mem={}kb, vmem={}kb, wall_time={:.3f}s'.format(
          cpu_time, proc_info.fMemResident, proc_info.fMemVirtual, wall_time))
    return None

def process_uproot_array(source):
    result = np.array(source)
    if (result.dtype == np.dtype('O')) and (isinstance(result[0], bytes)):
        result = result.astype(str)
    return result

def get_data_size(data):
    if isinstance(data, (tuple, list, np.ndarray)):
        return len(data)
    else:
        return 1

def uproot_to_dict(file):
    result = {}
    for tree in file.values():
        tree_name = tree.name.decode('utf-8')
        result[tree_name] = {}
        for branch in tree.values():
            branch_name = branch.name.decode('utf-8')
            arr = process_uproot_array(branch.array())
            value = arr[0] if len(arr) == 1 else arr
            result[tree_name][branch_name] = value    
    return result
    
def get_leaf_type_str(data):
    type_str = None
    data_type = None

    if isinstance(data, (tuple, list, np.ndarray)):
        data_type = type(data[0])
        if data_type in string_types:
            return None
    elif isinstance(data, string_types):
        return None
    else:
        data_type = type(data)
        
    if data_type in [float, np.float64]:
        type_str = 'D'
    elif data_type in [int, np.int64]:
        type_str = 'L'
    elif data_type in [np.int32]:
        type_str = 'I'
    elif data_type in [np.uint32]:
        type_str = 'i'        
    elif data_type in [bool, np.bool_]:
        type_str = 'O'        
    elif data_type in [np.float32]:
        type_str = 'F'
    elif data_type in [np.int8]:
        type_str = 'B'
    elif data_type in [np.uint8]:
        type_str = 'b'
    else:
        raise ValueError('cannot infer leaf type for {}'.format(data_type))
    return type_str

        
        
def fill_branch(tree, values_dict):
    buffer = {}
    leaf_types = {}
    data_sizes = list(set([get_data_size(v) for k,v in values_dict.items()]))
    if len(data_sizes) > 1:
        raise ValueError('cannot fill tree with values of inconsistent entry sizes')
    data_size = data_sizes[0]
    if data_size == 1:
        for k, v in values_dict.items():
            leaf_type_str = get_leaf_type_str(v)
            leaf_types[k] = leaf_type_str
            if leaf_type_str is not None:
                buffer[k] = np.array([v], dtype=type(v))
                tree.Branch(k, buffer[k], '{}/{}'.format(k, leaf_type_str))
            else:
                buffer[k] = ROOT.std.string(v)
                tree.Branch(k, buffer[k])
        tree.Fill()
    else:
        for k, v in values_dict.items():
            leaf_type_str = get_leaf_type_str(v)
            leaf_types[k] = leaf_type_str
            if leaf_type_str is not None:
                buffer[k] = np.array([v[0]], dtype=type(v[0]))
                tree.Branch(k, buffer[k], '{}/{}'.format(k, leaf_type_str))
            else:
                buffer[k] = ROOT.std.string(v[0])
                tree.Branch(k, buffer[k])
        for i in range(data_size):
            for k, v in values_dict.items():
                if leaf_types[k] is not None:
                    buffer[k][0] = v[i]
                else:
                    buffer[k].replace(0, ROOT.std.string.npos, v[i])
            tree.Fill()
    


from typing import Dict, Optional, Union, List

import pandas as pd
from cycler import cycler

from quickstats.plots import AbstractPlot, QUICKSTATS_PALETTES
from quickstats.utils.common_utils import combine_dict

class MultiDataPlot(AbstractPlot):
    
    COLOR_CYCLER = cycler(color=QUICKSTATS_PALETTES['darklines'])
    
    def __init__(self, data_map:Union[pd.DataFrame, Dict[str, pd.DataFrame]],
                 label_map:Optional[Dict]=None,
                 styles_map:Optional[Dict]=None,
                 color_cycler=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None):
        
        self.data_map = data_map
        self.label_map = label_map
        self.styles_map = styles_map
        
        if color_cycler is not None:
            self.color_cycler = color_cycler
        else:
            self.color_cycler = self.COLOR_CYCLER
            
        super().__init__(styles=styles, 
                         analysis_label_options=analysis_label_options)            
            
    def get_default_legend_order(self):
        if not isinstance(self.data_map, dict):
            return []
        else:
            return list(self.data_map)
    
    def draw_frame(self, **kwargs):
        ax = super().draw_frame(**kwargs)
        ax.set_prop_cycle(self.color_cycler)
        return ax
    


from quickstats.components.abstract_object import AbstractObject
from quickstats.components.extended_minimizer import ExtendedMinimizer
from quickstats.components.extended_model import ExtendedModel
from quickstats.components.analysis_object import AnalysisObject
from quickstats.components.asimov_generator import AsimovGenerator
from quickstats.components.asimov_generator import AsimovType
from quickstats.components.nuisance_parameter_pull import NuisanceParameterPull
from quickstats.components.nuisance_parameter_harmonizer import NuisanceParameterHarmonizer
from quickstats.components.asymptotic_cls import AsymptoticCLs
from quickstats.components.likelihood import Likelihood
from quickstats.components.pvalue_toys import PValueToys
from quickstats.components.toy_limit_calculator import ToyLimitCalculator
from quickstats.components.analysis_base import AnalysisBase
from quickstats.components.param_parser import ParamParser
from quickstats.components.roo_inspector import RooInspector


import click

import fnmatch
from quickstats.components import ExtendedModel
    
@click.command(name='inspect_ws')
@click.option('-i', '--input_file', required=True, help='Path to the input workspace file')
@click.option('-w', '--workspace', 'ws_name', default=None, help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, help='Name of model config. Auto-detect by default.')
@click.option('-o', '--output_file', default=None, help='Export output to text file. If None, no output is saved.')
@click.option('--items', default=None, help='\b\nItems to include in the summary(separated by commas). Choose from\n'
                   '\b workspace\n'
                   '\b dataset\n'
                   '\b category\n'
                   '\b poi\n'
                   '\b nuisance_parameter\n'
                   '\b global_observable\n')
@click.option('--include', 'include_patterns', default=None, 
              help='Match variable names with given patterns (separated by commas).')
@click.option('--exclude', 'exclude_patterns', default=None,
              help='Exclude variable names with given patterns (separated by commas).')
@click.option('--detailed/--name-only', default=True, show_default=True,
              help='Include detailed variable properties or just the variable name in the summary.')
def inspect_ws(input_file, ws_name=None, mc_name=None, output_file=None, items=None,
               include_patterns=None, exclude_patterns=None, detailed=True):
    '''
        Inspect workspace attributes
    '''
    model = ExtendedModel(input_file, ws_name=ws_name, mc_name=mc_name, verbosity="ERROR", data_name=None)
    items = items.split(",") if items is not None else None
    include_patterns = include_patterns.split(",") if include_patterns is not None else None
    exclude_patterns = exclude_patterns.split(",") if exclude_patterns is not None else None
    model.print_summary(items=items, save_as=output_file, detailed=detailed,
                        include_patterns=include_patterns, exclude_patterns=exclude_patterns)


import os
import math
import time
import json
from itertools import repeat
from typing import Optional, Dict, List, Union

import numpy as np
import ROOT
from quickstats.components import ExtendedModel, ExtendedMinimizer
from quickstats.components.numerics import approx_n_digit, str_encode_value, str_decode_value
from quickstats.utils import common_utils
from quickstats.concurrent.logging import standard_log

def load_cache(fname:str):
    if (fname is not None) and os.path.exists(fname):
        with open(fname, 'r') as file:
            data = json.load(file)
        return data
    else:
        return None

from quickstats.components import AnalysisObject
from quickstats.utils.common_utils import parse_config

class Likelihood(AnalysisObject):
    
    def __init__(self, filename:str, poi_name:Optional[str]=None,
                 data_name:str='combData', 
                 config:Optional[Union[Dict, str]]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):

        config = parse_config(config)
        config['filename']  = filename
        config['poi_name']  = poi_name
        config['data_name'] = data_name
        config['verbosity'] = verbosity
        self._inherit_init(super().__init__, **config)
        
        self.model.workspace.saveSnapshot("initialSnapshot", self.model.workspace.allVars())
        
    def nll_fit(self, poi_val:Optional[float]=None, mode:int=2,
                snapshot_name:Optional[str]="initialSnapshot",
                ):
        if poi_val is None:
            poi_val = self.poi.getVal()
        
        nll_uncond = None
        nll_cond   = None
        muhat      = None
        start_time = time.time()
        # unconditional nll
        if mode in [0, 1]:
            self.load_snapshot(snapshot_name)
            for poi in self.model.pois:
                poi.setConstant(1)
            self.poi.setConstant(0)
            self.poi.setVal(poi_val)
            fit_status_uncond = self.minimizer.minimize()
            muhat     = self.poi.getVal()
            muhat_errolo = self.poi.getErrorLo()
            muhat_errohi = self.poi.getErrorHi()
            nll_uncond = self.minimizer.nll.getVal()
        fit_time_uncond = time.time() - start_time
        # conditional nll
        if mode in [0, 2]:
            self.load_snapshot(snapshot_name)
            for poi in self.model.pois:
                poi.setConstant(1)            
            self.poi.setConstant(1)
            self.poi.setVal(poi_val)
            fit_status_cond = self.minimizer.minimize()
            nll_cond = self.minimizer.nll.getVal()
        fit_time_cond = time.time() - fit_time_uncond - start_time
        
        fit_result = {}
        self.stdout.info("INFO: NLL evaluation completed with")
        if mode in [0, 1]:
            self.stdout.info("\tmu hat = {:.5f}\n\tuncond NLL = {:.5f}".format(muhat, nll_uncond))
            fit_result['uncond_fit'] = {
                'muhat': muhat,
                'muhat_errlo': muhat_errolo,
                'muhat_errhi': muhat_errohi,
                'nll': nll_uncond,
                'status': fit_status_uncond,
                'time': fit_time_uncond
            }
        if mode in [0, 2]:
            self.stdout.info("\tmu = {:.5f}".format(poi_val))
            self.stdout.info("\tcond NLL = {:.5f}".format(nll_cond))
            fit_result['cond_fit'] = {
                'mu': poi_val,
                'nll': nll_cond,
                'status': fit_status_cond,
                'time': fit_time_cond
            }
        if mode == 0:
            pnll = nll_cond - nll_uncond
            qmu  = 2*pnll
            self.stdout.info("\tPNLL = {:.5f}".format(nll_cond - nll_uncond))
            fit_result['pnll'] = pnll
            fit_result['qmu']  = qmu
            
            if (qmu >= 0):
                significance = math.sqrt(qmu)
                pvalue = 1 - ROOT.Math.normal_cdf(significance, 1, poi_val)
                fit_result['significance'] = significance
                fit_result['pvalue'] = pvalue
            else:
                fit_result['significance'] = None
                fit_result['pvalue'] = None
                
        self.stdout.info("\ttime (uncond_fit, cond_fit) = {:.3f}, {:.3f}".format(fit_time_uncond, fit_time_cond))
        fit_result['total_time'] = fit_time_uncond + fit_time_cond
        
        return fit_result
        
    def evaluate_nll(self, poi_val:Optional[float]=None, mode:int=2,
                     snapshot_name:Optional[str]="initialSnapshot"):
        """Evalute post-fit NLL value
        
        Arguments:
            poi_val: (Optional) float
                Value of parameter of interest during fit. If not specified, the original value is used.
            mode: int
                Evaluation mode. Choose from:
                    0: Evaluate nll_mu - nll_muhat
                    1: Evaluate unconditional NLL (nll_muhat)
                    2: Evaluate conditional NLL (nll_mu)
            snapshot_name: (Optional) str
                Name of snapshot to load before fitting.
        """
        fit_result = self.nll_fit(poi_val=poi_val, mode=mode, snapshot_name=snapshot_name)
        
        if mode == 0:
            return fit_result['pnll']
        if mode == 1:
            return fit_result['uncond_fit']['nll']
        if mode == 2:
            return fit_result['cond_fit']['nll']
    
def evaluate_nll(filename:str, poi_val:Optional[float]=None, poi_name:str=None, mode:int=1,
                 ws_name:Optional[str]=None, mc_name:Optional[str]=None, data_name:str='combData', 
                 snapshot_name:str=None, profile_param:str="", fix_param:str="", 
                 constrain_nuis:bool=True, minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', 
                 num_cpu:int=1, binned_likelihood:bool=True, eps:float=1.0, strategy:int=1, 
                 optimize:int=2, offset:bool=True, fix_cache:bool=True, fix_multi:bool=True, 
                 max_calls:int=-1, max_iters:int=-1, print_level:int=-1, verbosity:int="WARNING",  
                 outname:str=None, cache:bool=False, detailed_output:bool=False, save_log:bool=False):
    if cache:
        cached_result = load_cache(outname)
        if cached_result is not None:
            poi_name, nll = cached_result['poi'], cached_result['nll']
            if not save_log:
                print('INFO: Found NLL cache from "{}"'.format(outname))
                if poi_val is None:
                    print('INFO: Cached unconditional NLL for POI "{}": {}'.format(poi_name, nll))
                else:
                    print('INFO: Cached NLL for POI "{}" at {:.2f}: {}'.format(poi_name, poi_val, nll))
            return nll
    
    config = {
        'ws_name' : ws_name,
        'mc_name' : mc_name,
        'snapshot_name': snapshot_name,
        'profile_param': profile_param,
        'fix_param': fix_param,
        'constrain_nuis': constrain_nuis,
        'minimizer_type': minimizer_type,
        'minimizer_algo': minimizer_algo,
        'num_cpu': num_cpu,
        'binned_likelihood': binned_likelihood,
        'eps': eps,
        'strategy': strategy,
        'optimize': optimize,
        'offset': offset,
        'fix_cache': fix_cache,
        'fix_multi': fix_multi,
        'max_calls': max_calls,
        'max_iters': max_iters,
        'print_level': print_level,
        'verbosity': verbosity,
    }
    if save_log and (outname is not None):
        log_path = os.path.splitext(outname)[0] + ".log"
    else:
        log_path = None
        
    with standard_log(log_path) as logger:
        if poi_val is None:
            print(f'INFO: Evaluating unconditional NLL for POI "{poi_name}"')
        else:
            print(f'INFO: Evaluating conditional NLL for POI "{poi_name}" at mu = {poi_val}')
        start = time.time()
        likelihood = Likelihood(filename=filename, poi_name=poi_name, data_name=data_name, config=config,
                                verbosity=verbosity)
        poi_name = likelihood.poi.GetName()
        nll = likelihood.evaluate_nll(poi_val, mode=mode)
        end = time.time()
        results = {
            'nll': nll,
            'poi': poi_name,
            'offset': int(offset),
            'constrain_nuis': int(constrain_nuis),
            'poi_value': poi_val
        }
        if poi_val is None:
            poi_val = likelihood.poi.getVal()
            results['poi_bestfit'] = poi_val
        results['time'] = end-start

        # save results
        if outname is not None:
            with open(outname, 'w') as outfile:
                json.dump(results, outfile)
            print('INFO: Saved NLL result to {}'.format(outname))

    if detailed_output:
        return results
    return nll

def scan_nll(filename:str, poi_min:float, poi_max:float, poi_step:float, poi_name:Optional[str]=None, 
             cache:bool=True, outname:str="{poi_name}.json", outdir:str='output', ws_name:Optional[str]=None, 
             mc_name:Optional[str]=None, data_name:str='combData', snapshot_name:str=None, 
             profile_param:str="", fix_param:str="", constrain_nuis:bool=True, minimizer_type:str='Minuit2',
             minimizer_algo:str='Migrad', num_cpu:int=1, binned_likelihood:bool=True, eps:float=1.0, 
             strategy:int=0, fix_cache:bool=True, fix_multi:bool=True, verbosity:int="WARNING", 
             print_level:int=-1, max_calls:int=-1, max_iters:int=-1, optimize:int=2, offset:bool=True, 
             parallel:int=-1, save_log:bool=True, **kwargs):
    start_time = time.time()
    points     = np.arange(poi_min, poi_max+poi_step, poi_step)
    points     = np.concatenate(([None], points))
    modes      = [1] + [2]*(len(points)-1)
    # try to get the number of significant digits of poi_step
    sd = approx_n_digit(poi_step)
    if poi_name is None:
        poi_name = ExtendedModel.get_poi_names(filename, ws_name, mc_name)[0]
        print('INFO: POI not given, default as "{}".'.format(poi_name))
    outname_noext = os.path.splitext(outname)[0]
    # save cache file as outdir/cache/outname_{str_encoded_poi_value}.json
    outnames = ["{}_{}.json".format(outname_noext.format(poi_name=poi_name), "uncond" if point is None \
                else str_encode_value(point, sd)) for point in points]
    cachedir = os.path.join(outdir, "cache")
    outnames = [os.path.join(cachedir, name) for name in outnames]
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    if not os.path.exists(cachedir):
        os.makedirs(cachedir)

    arguments = (repeat(filename), points, repeat(poi_name), modes, repeat(ws_name),repeat(mc_name), 
                 repeat(data_name), repeat(snapshot_name), repeat(profile_param), repeat(fix_param), 
                 repeat(constrain_nuis), repeat(minimizer_type), repeat(minimizer_algo), repeat(num_cpu),
                 repeat(binned_likelihood), repeat(eps), repeat(strategy), repeat(optimize), repeat(offset),
                 repeat(fix_cache), repeat(fix_multi), repeat(max_calls), repeat(max_iters), 
                 repeat(print_level), repeat(verbosity), outnames, repeat(cache), repeat(False),
                 repeat(save_log))
    
    results = common_utils.execute_multi_tasks(evaluate_nll, *arguments, parallel=parallel)

    uncond_nll = results[0]
    data = {'mu': [], 'nll':[], 'qmu':[]}
    for mu, nll in zip(points, results):
        data['mu'].append(mu)
        data['nll'].append(nll)
        data['qmu'].append(2*(nll-uncond_nll))
    final_outname = os.path.join(outdir, outname.format(poi_name=poi_name))
    with open(final_outname, 'w') as outfile:
        json.dump(data, outfile, indent=3)
    end_time = time.time()
    print('INFO: All jobs have finished. Total time taken: {:.3f} s'.format(end_time-start_time))  


from typing import Optional, Union, List, Dict

import numpy as np
import ctypes
import cppyy
from cppyy.gbl.std import vector

import ROOT
import ROOT.RooStats as RS

def get_null_distribution(htr:ROOT.RooStats.HypoTestResult)->np.ndarray:
    return np.array(htr.GetNullDistribution().GetSamplingDistribution().data())

def get_alt_distribution(htr:ROOT.RooStats.HypoTestResult)->np.ndarray:
    return np.array(htr.GetAltDistribution().GetSamplingDistribution().data())

def get_merged_null_distribution(htr_list:List[ROOT.RooStats.HypoTestResult])->np.ndarray:
    return np.sort(np.concatenate(tuple([get_null_distribution(htr) for htr in htr_list])))

def get_merged_alt_distribution(htr_list:List[ROOT.RooStats.HypoTestResult])->np.ndarray:
    return np.sort(np.concatenate(tuple([get_alt_distribution(htr) for htr in htr_list])))

def process_hypotest_results(htr_list:List[RS.HypoTestResult],
                             remove_unconverged_toys:bool=True, 
                             use_qmu:bool=False,
                             do_invert:bool=True):
    if not isinstance(htr_list, list):
        htr_list = [htr_list]
    teststat_list = [htr.GetTestStatisticData() for htr in htr_list]
    if len(set(teststat_list)) > 1:
        raise RuntimeError("inconsistent test statistic values among hypotest results")
    teststat = teststat_list[0]
    if (teststat < 0.):
        print("WARNING: HypoTestResult has negative test statistic indicating a failed fit.")
    null_dist = get_merged_null_distribution(htr_list)
    alt_dist = get_merged_alt_distribution(htr_list)

    if remove_unconverged_toys:
        null_dist = null_dist[null_dist >= 0.]
        alt_dist  = alt_dist[alt_dist >= 0.]

    if use_qmu:
        null_dist *= 2
        alt_dist  *= 2
        teststat  *= 2
    null_dist_vec = vector['double'](null_dist)
    alt_dist_vec  = vector['double'](alt_dist)
    null_sampling_dist = RS.SamplingDistribution("null_dist", "", null_dist_vec)
    alt_sampling_dist  = RS.SamplingDistribution("alt_dist", "", alt_dist_vec)
    new_htr = htr_list[0].Clone()
    new_htr.SetNullDistribution(null_sampling_dist)
    new_htr.SetAltDistribution(alt_sampling_dist)
    new_htr.SetTestStatisticData(teststat)
    if do_invert:
        new_htr.SetBackgroundAsAlt()
    return new_htr

def merge_toy_results(results:List[Union[RS.HypoTestInverterResult,
                      RS.HypoTestResult]],
                      poi:Optional[ROOT.RooRealVar]=None,
                      interpolation_method:int=RS.HypoTestInverterResult.kLinear,
                      remove_mu_with_failed_teststat:bool=True,
                      remove_unconverged_toys:bool=True, 
                      use_qmu:bool=True,
                      do_invert:bool=True,
                      silent:bool=False):
    hypotest_inverter_results = [r for r in results if r.ClassName() == 'RooStats::HypoTestInverterResult']
    hypotest_results          = [r for r in results if r.ClassName() == 'RooStats::HypoTestResult']

    if (not hypotest_inverter_results) and (not hypotest_results):
        raise ValueError("toy results must be either instance of "
                         "RooStats::HypoTestInverterResult or RooStats::HypoTestResult")
    if poi is None:
        if len(hypotest_inverter_results) > 0:
            base_result = hypotest_inverter_results[0]
            poi = base_result.GetParameters().first()
        else:
            base_result = hypotest_results[0]
            poi_name    = '_'.join(base_result.GetName().split('_')[:-1])
            poi         = ROOT.RooRealVar(poi_name, poi_name, 0)
            print("WARNING: POI information not given. New POI variable will be constructed "
                  " using name inferred from toy result (\"{}\"={}[{}, {}]).".format(
                      poi_name, poi.getVal(), poi.getRange()[0], poi.getRange()[1]))
    merged_result = RS.HypoTestInverterResult("merged_result", poi, 0.95)
    merged_result.SetConfidenceLevel(0.95)
    merged_result.UseCLs()
    merged_result.SetInterpolationOption(interpolation_method)
    if not silent:
        if interpolation_method == RS.HypoTestInverterResult.kLinear:
            interpolation_str = "Linear"
        elif interpolation_method == RS.HypoTestInverterResult.kSpline:
            interpolation_str = "Spline"        
        print("INFO: Constructing HypoTestInverterResult with the following settings")
        print("                  CL: 0.95")
        print("             Use CLs: True")
        print("Interpolation Method: {}".format(interpolation_str))
    temp_results = {}
    for result in hypotest_inverter_results:
        r_poi = result.GetParameters().first()
        if r_poi.GetName() != poi.GetName():
            raise RuntimeError("inconsistent POI used across toy results")
        for i in range(result.ArraySize()):
            htr = result.GetResult(i)
            mu = result.GetXValue(i)
            if mu not in temp_results:
                temp_results[mu] = []
            temp_results[mu].append(htr)
        
    for result in hypotest_results:
        try:
            mu = float(result.GetName().split('_')[-1])
        except:
            raise RuntimeError("failed to extract mu value from HypoTestResult name "
                               "(expect <poi_name>_<mu_value>)")
        if mu not in temp_results:
            temp_results[mu] = []
        temp_results[mu].append(result)

    for mu, htr_list in temp_results.items():
        teststat_list = [htr.GetTestStatisticData() for htr in htr_list]
        if any(teststat < 0 for teststat in teststat_list) and remove_mu_with_failed_teststat:
            print(f"WARNING: Removed results from mu = {mu} due to negative teststat value")
            continue
        new_htr = process_hypotest_results(htr_list,
                  remove_unconverged_toys=remove_unconverged_toys, 
                  use_qmu=use_qmu,
                  do_invert=do_invert)
        merged_result.Add(mu, new_htr)
    return merged_result

def get_array_quantiles_deprecated(array):
    import cppyy.ll
    q = cppyy.ll.array_new['double'](7)
    p = cppyy.ll.array_new['double'](7)
    p[0] = ROOT.Math.normal_cdf(-2)
    p[1] = ROOT.Math.normal_cdf(-1)
    p[2] = 0.5
    p[3] = ROOT.Math.normal_cdf(1)
    p[4] = ROOT.Math.normal_cdf(2)
    ROOT.TMath.Quantiles(len(array), 5, array, q, p, False)
    quantiles = {
        -2 : q[0],
        -1 : q[1],
         0 : q[2],
        +1 : q[3],
        +2 : q[4]
    }
    # free memory
    cppyy.ll.array_delete(p)
    cppyy.ll.array_delete(q)
    return quantiles

def get_array_quantiles(array):
    q = [ROOT.Math.normal_cdf(-2), ROOT.Math.normal_cdf(-1), 0.5, ROOT.Math.normal_cdf(1), ROOT.Math.normal_cdf(2)]
    quantiles = np.quantile(array, q)
    quantiles_dict = {
        -2 : quantiles[0],
        -1 : quantiles[1],
         0 : quantiles[2],
        +1 : quantiles[3],
        +2 : quantiles[4]
    }
    return quantiles_dict


def get_hypotest_data(htr):
    data = {}
    data['null'] = {
        'dist': htr.GetNullDistribution()
    }
    data['alt'] = {
        'dist': htr.GetAltDistribution()
    }
    for dtype in data:
        dist = data[dtype]['dist']
        data[dtype]['data'] = np.array(dist.GetSamplingDistribution().data())
        data[dtype]['weight'] = np.array(dist.GetSampleWeights().data())        
        data[dtype]['size'] = len(data[dtype]['data'])
        data[dtype]['xmin'] = np.min(data[dtype]['data'])
        data[dtype]['xmax'] = np.max(data[dtype]['data'])
    data['observed'] = {
        'teststat': htr.GetTestStatisticData(),
        'CLsplusb': htr.CLsplusb(),
        'CLsplusbError': htr.CLsplusbError(),
        'CLb': htr.CLb(),
        'CLbError': htr.CLbError(),
        'CLs': htr.CLs(),
        'CLsError': htr.CLsError(),
    }
    data['expected'] = {}
    # calculate expected intervals for teststat
    alt_data = data['alt']['dist'].GetSamplingDistribution().data()
    data['expected']['teststat'] = get_array_quantiles(alt_data)
    qmu_exp = data['expected']['teststat'][0]
    perror = ctypes.c_double()
    
    data['expected']['CLsplusb'] = data['null']['dist'].IntegralAndError(perror, qmu_exp, 
                                                                         ROOT.RooNumber.infinity(), 
                                                                         True, True, True)
    data['expected']['CLsplusbError'] = perror.value
    
    data['expected']['CLb'] = data['alt']['dist'].IntegralAndError(perror, qmu_exp, 
                                                                   ROOT.RooNumber.infinity(), 
                                                                   True, True, True)
    data['expected']['CLbError'] = perror.value
    
    if data['expected']['CLb'] == 0.:
        data['expected']['CLs'] = -1.
        data['expected']['CLsError'] = -1.
    else:
        data['expected']['CLs'] = data['expected']['CLsplusb'] / data['expected']['CLb']
        cl_b_err2 = pow(data['expected']['CLbError'], 2)
        cl_sb_err2 = pow(data['expected']['CLsplusbError'], 2)
        data['expected']['CLsError'] = np.sqrt(cl_sb_err2 + cl_b_err2 * pow(data['expected']['CLs'], 2)) / data['expected']['CLb']

    return data


from typing import Optional, List, Dict, Union
import os
import glob
import json
import fnmatch

import ROOT

from quickstats.components import AbstractObject

class RooInspector(AbstractObject):
    def __init__(self, tree_name:str, file_expr:Union[str, List[str]], 
                 filter_expr:Optional[str]=None,
                 verbosity:Optional[Union[int, str]]="INFO"):
        super().__init__(verbosity=verbosity)
        self.tree_name = tree_name
        self.fnames = self._parse_file_expr(file_expr)
        self.initialise(filter_expr=filter_expr)
        
    def _parse_file_expr(self, file_expr:Union[str, List[str]])->List[str]:
        fnames = []
        if isinstance(file_expr, str):
            return self._parse_file_expr([file_expr])
        else:
            for expr in file_expr:
                if os.path.isdir(expr):
                    file_expr = os.path.join(expr, "*.root")
                fnames_i = glob.glob(expr)
                if not fnames_i:
                    self.stdout.warning(f"WARNING: No root files found matching the expression {expr}")
                fnames.append(fnames_i)
        if not fnames:
            raise RuntimeError("no root files found from the given file expression")
        return fnames
    
    def initialise(self, filter_expr:Optional[str]=None):
        self.rdf = ROOT.RDataFrame(self.tree_name, self.fnames)
        if filter_expr is not None:
            self.rdf = self.rdf.Filter(filter_expr)
        
    def get_column_names(self)->List[str]:
        column_names = [str(i) for i in self.rdf.GetColumnNames()]
        return column_names
    
    def get_column_types(self, column_names:List[str])->Dict[str,str]:
        all_column_names = self.get_column_names()
        invalid_column_names = set(column_names) - set(all_column_names)
        if len(invalid_column_names) > 0:
            raise RuntimeError("unknown column names: {}".format(",".join(invalid_column_names)))
        return {column_name:str(self.rdf.GetColumnType(column_name)) for column_name in column_names}
    
    def get_entries(self):
        return self.rdf.Count().GetValue()
    
    def print_summary(self, suppress_print:bool=False,
                      include_patterns:Optional[List]=None, exclude_patterns:Optional[List]=None,
                      save_as:Optional[str]=None):
        summary_str = ""
        nentries = self.get_entries()
        summary_str += f"Number of Events: {nentries}\n"
        column_names = self.get_column_names()
        if include_patterns is not None:
            column_names = [name for pattern in include_patterns for name in column_names if fnmatch.fnmatch(name, pattern)]
        if exclude_patterns is not None:
            column_names = [name for pattern in exclude_patterns for name in column_names if not fnmatch.fnmatch(name, pattern)]
        column_types = self.get_column_types(column_names)
        n_columns = len(column_types)
        summary_str += f"Columns of Interest ({n_columns}):\n"
        for cname, ctype in column_types.items():
            ctype_str = "(" + ctype + ")"
            summary_str += f"{ctype_str:<30}{cname}\n"
        if not suppress_print:
            self.stdout.info(summary_str)
        if save_as is not None:
            with open(save_as, "w") as f:
                f.write(summary_str)
        
        

    


import os
import click

from quickstats.concurrent.parameterised_asymptotic_cls import run_param_scan

@click.command(name='limit_scan')
@click.option('-i', '--input_dir', 'dirname', required=True, 
              help='Input directory containing the workspace files to process.')
@click.option('--file_expr', default="<filename[S]>", show_default=True,
              help='\b File name expression describing the external parameterisation.\n'
                   '\b Example: "<mass[F]>_kl_<klambda[P]>"\n'
                   '\b Refer to documentation for more information\n')
@click.option('--param_expr', default=None, show_default=True,
              help='\b Parameter name expression describing the internal parameterisation.\n'
                   '\b Example: "klambda=-10_10_0.2,k2v=1"\n'
                   '\b Refer to documentation for more information\n')
@click.option('--outdir', default='output', show_default=True,
              help='Output directory where cached limit files and merged limit file are saved.')
@click.option('-o', '--outname', default='limits.json', show_default=True,
              help='Name of output limit file (all parameter points merged).')
@click.option('--cache/--no-cache', default=True, show_default=True,
              help='Cache output of individual parameter point')
@click.option('--save-log/--no-log', default=True, show_default=True,
              help='Save a log file for each parameter point.')
@click.option('--save-summary/--no-summary', default=True, show_default=True,
              help='Save a summary file for each parameter point.')
@click.option('--parallel', type=int, default=-1, show_default=True,
              help='\b Parallelize job across the N workers.'
                   '\b Case  0: Jobs are run sequentially (for debugging).\n'
                   '\b Case -1: Jobs are run across N_CPU workers.\n')
@click.option('-p', '--poi', 'poi_name', default=None,
              help='Name of the parameter of interest (POI). If None, the first POI is used.')
@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
              help='Name of dataset')
@click.option('--blind/--unblind', 'do_blind', default=True, show_default=True,
              help='Blind/unblind analysis')
@click.option('--CL', 'CL', type=float, default=0.95, show_default=True,
              help='CL value to use')
@click.option('--precision', default=0.005, show_default=True,
              help='precision in mu that defines iterative cutoff')
@click.option('--do_tilde/--no_tilde', default=True, show_default=True,
              help='bound mu at zero if true and do the \tilde{q}_{mu} asymptotics')
@click.option('--predictive_fit/--no_predictive_fit', default=True, show_default=True,
              help='extrapolate best fit nuisance parameters based on previous fit results')
@click.option('--do_better_bands/--skip_better_bands', default=True, show_default=True,
              help='evaluate asymptotic CLs limit for various sigma bands')
@click.option('--better_negative_bands/--skip_better_negative_bands', default=False, show_default=True,
              help='evaluate asymptotic CLs limit for negative sigma bands')
@click.option('--binned/--unbinned', 'binned_likelihood', default=True, show_default=True,
              help='Binned likelihood')
@click.option('--save_summary/--skip_summary', default=True, show_default=True,
              help='Save summary information')
@click.option('-f', '--fix', 'fix_param', default="", show_default=True,
              help='Parameters to fix')
@click.option('-r', '--profile', 'profile_param', default="", show_default=True,
              help='Parameters to profile')
@click.option('-w', '--workspace', 'ws_name', default=None, show_default=True,
              help='Name of workspace. Auto-detect by default.')
@click.option('-m', '--model_config', 'mc_name', default=None, show_default=True,
              help='Name of model config. Auto-detect by default.')
@click.option('-s', '--snapshot', 'snapshot_name', default=None, show_default=True,
              help='Name of initial snapshot')
@click.option('-t', '--minimizer_type', default="Minuit2", show_default=True,
              help='Minimizer type')
@click.option('-a', '--minimizer_algo', default="Migrad", show_default=True,
              help='Minimizer algorithm')
@click.option('-e', '--eps', type=float, default=1.0, show_default=True,
              help='Convergence criterium')
@click.option('--strategy', type=int, default=1, show_default=True,
              help='Default minimization strategy')
@click.option('--print_level', type=int, default=-1, show_default=True,
              help='Minimizer print level')
@click.option('--timer/--no_timer', default=False, show_default=True,
              help='Enable minimizer timer')
@click.option('-c', '--num_cpu', type=int, default=1, show_default=True,
              help='Number of CPUs to use per parameter')
@click.option('--offset/--no-offset', default=True, show_default=True,
              help='Offset likelihood')
@click.option('--optimize', type=int, default=2, show_default=True,
              help='Optimize constant terms')
@click.option('--fix-cache/--no-fix-cache', default=True, show_default=True,
              help='Fix StarMomentMorph cache')
@click.option('--fix-multi/--no-fix-cache',  default=True, show_default=True,
              help='Fix MultiPdf level 2')
@click.option('--max_calls', type=int, default=-1, show_default=True,
              help='Maximum number of function calls')
@click.option('--max_iters', type=int, default=-1, show_default=True,
              help='Maximum number of Minuit iterations')
@click.option('--batch_mode/--no-batch', default=False, show_default=True,
              help='Batch mode when evaluating likelihood')
@click.option('--int_bin_precision', type=float, default=-1., show_default=True,
              help='Integrate the PDF over the bins instead of using the probability '
                   'density at the bin centre')
@click.option('--constrain/--no-constrain', 'constrain_nuis', default=True, show_default=True,
              help='Use constrained NLL')
@click.option('-v', '--verbosity',  default="INFO", show_default=True,
              help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
def limit_scan(**kwargs):
    """
    Evaluate a set of parmeterised asymptotic cls limits
    """
    _kwargs = {}
    for arg_name in ["dirname", "file_expr", "param_expr", "outdir", 
                     "outname", "cache", "save_log", "save_summary",
                     "parallel"]:
        _kwargs[arg_name] = kwargs.pop(arg_name)
    _kwargs['config'] = kwargs
    run_param_scan(**_kwargs)


EXTRA_COLORS = {
    "paper:bg": "#eeeeee",
    "paper:fg": "#444444",
    "paper:bgAlt": "#e4e4e4",
    "paper:red": "#af0000",
    "paper:green": "#008700",
    "paper:blue": "#005f87",
    "paper:yellow": "#afaf00",
    "paper:orange": "#d75f00",
    "paper:pink": "#d70087",
    "paper:purple": "#8700af",
    "paper:lightBlue": "#0087af",
    "paper:olive": "#5f7800",
    "on:bg": "#1b2b34",
    "on:fg": "#cdd3de",
    "on:bgAlt": "#343d46",
    "on:fgAlt": "#d8dee9",
    "on:red": "#ec5f67",
    "on:orange": "#f99157",
    "on:yellow": "#fac863",
    "on:green": "#99c794",
    "on:cyan": "#5fb3b3",
    "on:blue": "#6699cc",
    "on:pink": "#c594c5",
    "on:brown": "#ab7967",
    "series:cyan": "#54c9d1",
    "series:orange": "#eca89a",
    "series:blue": "#95bced",
    "series:olive": "#ceb776",
    "series:purple": "#d3a9ea",
    "series:green": "#9bc57f",
    "series:pink": "#f0a1ca",
    "series:turquoise": "#5fcbaa",
    "atlas:onesigma": "#00ff26",
    "atlas:twosigma": "#fbff1f",
    "series2:green": "#00ff26",
    "series2:yellow": "#fbff1f",
    "series2:blue": "#00a1e0",
    "series2:red": "#a30013",
    "series2:purple": "#5100c2",
    "hdbs:starcommandblue": "#047cbc",
    "hdbs:spacecadet": "#283044",
    "hdbs:mintcream": "#ebf5ee",
    "hdbs:outrageousorange": "#fa7e61",
    "hdbs:pictorialcarmine": "#ca1551",
    "hdbs:maroonX11": "#b8336a",
    "hh:darkpink": "#f2385a",
    "hh:darkblue": "#343844",
    "hh:medturquoise": "#36b1bf",
    "hh:lightturquoise": "#4ad9d9",
    "hh:offwhite": "#e9f1df",
    "hh:darkyellow": "#fdc536",
    "hh:darkgreen": "#125125",
    "transparent": "#ffffff00",
}

QUICKSTATS_PALETTES  = dict(
    darklines=["#000000", "#0019f5", "#377d22", "#ea3423", "#f2a93b", "#7f170e"]
)


import os
import sys
import copy
import json
from typing import Optional, Union, Dict, List
from itertools import repeat

import pandas as pd

from quickstats.components import AsymptoticCLs, ParamParser
from quickstats.concurrent.logging import standard_log
from quickstats.utils.common_utils import execute_multi_tasks

def run_param_point(filename:str, parameters:Optional[Dict]=None,
                    outname:str="limit.json", cache:bool=True,
                    save_log:bool=True, save_summary:bool=True,
                    config:Optional[Dict]=None):

    if parameters is not None:
        param_str = "(" + ", ".join([f"{param}={round(value, 8)}" for param, value in parameters.items()]) + ")"
    else:
        param_str = ""
    sys.stdout.write(f"INFO: Evaluating limit for the workspace {filename} {param_str}\n")

    if cache and os.path.exists(outname):
        sys.stdout.write(f"INFO: Cached limit output from {outname}\n")
        with open(outname, 'r') as f:
            limits = json.load(f)
            return limits

    if save_log:
        log_path = outname.replace(".json", ".log")
    else:
        log_path = None

    config['filename'] = filename

    asymptotic_cls = None
    with standard_log(log_path) as logger:
        asymptotic_cls = AsymptoticCLs(**config)
        asymptotic_cls.evaluate_limits()
        if outname is not None:
            asymptotic_cls.save(outname, summary=save_summary)
    if asymptotic_cls is None:
        return {}
    return asymptotic_cls.limits
    
def run_param_scan(dirname:str="", file_expr:Optional[str]=None, 
                   param_expr:Optional[str]=None, outdir:str="output",
                   outname:str="limits.json", cache:bool=True,
                   save_log:bool=True, save_summary:bool=True,
                   parallel:int=-1, config:Optional[Dict]=None):
    parser = ParamParser(file_expr, param_expr)
    param_points = parser.get_param_points(dirname)
    
    if config is None:
        config = {} 
    fix_param = config.get("fix_param", "")
    
    fnames     = []
    parameters = []
    cache_outnames = []
    configs    = []
    
    for point in param_points:
        int_params = point['internal_parameters']
        ext_params = point['external_parameters']
        if set(int_params) & set(ext_params):
            raise RuntimeError("internal and external parameters are not mutually exclusive")
        if len(int_params) + len(ext_params) == 0:
            raise RuntimeError("no parameters to scan for")
        all_params = {**int_params, **ext_params}
        parameters.append(all_params)
        
        fname = point['filename']
        fnames.append(fname)
        
        point_config = copy.deepcopy(config)
        val_expr = parser.val_encode_parameters(int_params)
        fix_expr = []
        if fix_param:
            fix_expr.append(fix_param)
        if val_expr:
            fix_expr.append(val_expr)
        point_config['fix_param'] = ",".join(fix_expr)
        configs.append(point_config)

        str_encoded_name = parser.str_encode_parameters(all_params)
        cache_outname = os.path.join(outdir, f"{str_encoded_name}.json")
        cache_outnames.append(cache_outname)
        
    argument_list = (fnames, parameters, cache_outnames, repeat(cache), repeat(save_log),
                     repeat(save_summary), configs)
    
    if ((outname is not None) or cache) and not os.path.exists(outdir):
        os.makedirs(outdir, exist_ok=True)
        
    limit_result = execute_multi_tasks(run_param_point, *argument_list, parallel=parallel)
    
    final_result = []

    for fname, params, limit in zip(fnames, parameters, limit_result):
        if len(limit) == 0:
            param_str = parser.val_encode_parameters(params)
            raise RuntimeError(f'Job failed for the input "{fname}" ({param_str}). '
                               'Please check the log file for more details.')
        final_result.append({**params, **limit})
    final_result = pd.DataFrame(final_result).to_dict('list')
    
    if outname is not None:
        outpath = os.path.join(outdir, outname)
        with open(outpath, "w") as f:
            json.dump(final_result, f, indent=2)


from typing import Optional, Union, Dict, List

import matplotlib.patches as patches
import matplotlib.lines as lines
import pandas as pd
import ROOT

from quickstats.plots import AbstractPlot
from quickstats.plots.template import single_frame, parse_styles
from quickstats.utils.roostats_utils import get_array_quantiles

class HypoTestInverterPlot(AbstractPlot):
    
    COLOR_PALLETE = {
        '2sigma': '#FDC536',
        '1sigma': '#4AD9D9',
        'CLs': 'r',
        'CLb': 'k',
        'CLsplusb': 'b'
    }
    
    def __init__(self, result, nsig:int=2, 
                 color_pallete:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]='limit',
                 analysis_label_options:Optional[Dict]=None):
        self.result = result
        
        super().__init__(color_pallete=color_pallete,
                         styles=styles,
                         analysis_label_options=analysis_label_options)
        
    def get_expected_CLs_data(self):
        n_entries = self.result.ArraySize()
        data = {}
        for i in range(n_entries):
            s = self.result.GetExpectedPValueDist(i)
            if not s:
                continue
            values = s.GetSamplingDistribution().data()
            mu = self.result.GetXValue(i)
            data[mu] = get_array_quantiles(values)
        return data
        
    def draw_expected(self, ax, nsig:int=2):
        data = self.get_expected_CLs_data()
        df = pd.DataFrame(data).transpose().sort_index()
        mu_values = df.index.values
        n2sigma = df[-2].values
        n1sigma = df[-1].values
        median  = df[0].values
        p1sigma = df[1].values
        p2sigma = df[2].values

        if nsig >= 2:
            ax.fill_between(mu_values, n2sigma, p2sigma, facecolor=self.color_pallete['2sigma'], 
                            label='Expected CLs $\pm 2\sigma$')
        if nsig >= 1:
            ax.fill_between(mu_values, n1sigma, p1sigma, facecolor=self.color_pallete['1sigma'], 
                            label='Expected CLs $\pm 1\sigma$')
        ax.plot(mu_values, median, 'k--', label='Expected CLs Median')            
    
    def get_observed_CL_data(self):
        n_entries = self.result.ArraySize()
        data = {}
        for i in range(n_entries):
            mu = self.result.GetXValue(i)
            data[mu] = {
                'CLs'          : self.result.CLs(i),
                'CLb'          : self.result.CLb(i),
                'CLsplusb'     : self.result.CLsplusb(i),
                'CLsError'     : self.result.CLsError(i),
                'CLbError'     : self.result.CLbError(i),
                'CLsplusbError': self.result.CLsplusbError(i),
            }
        return data
    
    def draw_observed(self, ax, draw_CLs:bool=True, draw_CLb:bool=True, 
                      draw_CLsplusb:bool=True):
        data = self.get_observed_CL_data()
        df = pd.DataFrame(data).transpose().sort_index()
        mu_values = df.index.values
        CLs = df['CLs'].values
        CLb = df['CLb'].values
        CLsplusb = df['CLsplusb'].values
        CLsError = df['CLsError'].values
        CLbError = df['CLbError'].values
        CLsplusbError = df['CLsplusbError'].values
        if draw_CLs:
            ax.errorbar(mu_values, CLs, yerr=CLsError, **self.styles['errorbar'], 
                        color=self.color_pallete['CLs'], linestyle='-', 
                        label='Observed CLs')
        if draw_CLb:
            ax.errorbar(mu_values, CLb, yerr=CLbError, **self.styles['errorbar'], 
                        color=self.color_pallete['CLb'], linestyle='-',
                        label='Observed CLb')
        if draw_CLsplusb:
            ax.errorbar(mu_values, CLsplusb, yerr=CLsplusbError, **self.styles['errorbar'], 
                        color=self.color_pallete['CLsplusb'], linestyle='dotted', 
                        label='Observed CLs+b')

    def draw(self, nsig:int=2, draw_CLs:bool=True, 
             draw_CLb:bool=True, draw_CLsplusb:bool=True,
             ymax:float=0.55, draw_CL:bool=True):
        if nsig < 0 or nsig > 2:
            raise ValueError("nsig must be between 0 and 2")
            
        ax = self.draw_frame()
        
        self.draw_expected(ax, nsig=nsig)
        self.draw_observed(ax, draw_CLs=draw_CLs,
                           draw_CLb=draw_CLb,
                           draw_CLsplusb=draw_CLsplusb)
        ax.set_xlabel("$\mu$", **self.styles['xlabel'])
        ax.set_ylabel("p value", **self.styles['ylabel'])
        
        ax.set_ylim(-0.01, ymax)
        n_entries = self.result.ArraySize()
        mu_list = []
        for i in range(n_entries):
            mu_list.append(self.result.GetXValue(i))
        ax.set_xlim(min(mu_list), max(mu_list))
        
        if draw_CL:
            ax.axhline(y=0.05, color='r', linestyle='-', linewidth=1)
        # border for the legend
        border_leg = patches.Rectangle((0, 0), 1, 1, facecolor = 'none', edgecolor = 'black', linewidth = 1)        
        # reorder legends
        handles, labels = ax.get_legend_handles_labels()
        if nsig == 0:
            handles = handles[1:] + [handles[0]]
            labels = labels[1:] + [labels[0]]
        elif nsig == 1:
            handles[1].set_linewidth(1.0)
            handles = handles[2:] + [handles[0], (handles[1], border_leg)]
            labels = labels[2:] + [labels[0], labels[1]]
        elif nsig == 2:
            handles[1].set_linewidth(1.0)
            handles[2].set_linewidth(1.0)
            handles = handles[3:] + [handles[0], (handles[1], border_leg), (handles[2], border_leg)]
            labels = labels[3:] + [labels[0], labels[1], labels[2]]
        ax.legend(handles, labels, loc='upper right', frameon=False, **self.styles['legend'])
        return ax


from typing import Optional, Union, List, Dict
import os
import time
import math
import json
from itertools import repeat

from quickstats import semistaticmethod
from quickstats.components import AbstractObject, ExtendedModel
from quickstats.components.numerics import str_encode_value
from quickstats.utils.common_utils import execute_multi_tasks
from quickstats.utils.root_utils import load_macro
from quickstats.utils.roostats_utils import process_hypotest_results, merge_toy_results

import ROOT

class ToyLimitCalculator(AbstractObject):
    
    HYPOTEST_OPTIONS = {
        'remove_mu_with_failed_teststat': True,
        'remove_unconverged_toys': True,
        'use_qmu': False,
        'do_invert': True
    }
    
    def __init__(self, filename:str, data_name:str,
                 n_toys:int, seed:int=1234,
                 poi_name:Optional[str]=None,
                 poi_min:Optional[float]=None,
                 poi_max:Optional[float]=None,
                 minimizer_type:str="Minuit2", strategy:int=1,
                 eps:float=0.05, tolerance:float=1., print_level:int=-1,
                 offset:bool=False, do_coarse_scan:bool=False,
                 fix_param:str='', profile_param:str='',
                 snapshot_name:Optional[Union[List[str], str]]=None,                 
                 verbosity:Optional[Union[int, str]]=None,
                 hypotest_options:Optional[Dict]=None,
                 mode:str='toy',
                 use_custom_teststat_cls:bool=True):
        super().__init__(verbosity=verbosity)
        self.model_base = ExtendedModel(filename,
                                        data_name=data_name,
                                        snapshot_name=snapshot_name,
                                        verbosity="WARNING")
        
        self.inverter_verbose = print_level >= 0
        
        self.hypotest_options = self.HYPOTEST_OPTIONS
        if hypotest_options is not None:
            self.hypotest_options.update(hypotest_options)
        
        if fix_param:
            self.model_base.fix_parameters(fix_param)
        if profile_param:
            self.model_base.profile_parameters(profile_param)
        
        self.poi = self.model_base.get_poi(poi_name)
        self.poi.setConstant(False)
        poi_range = [self.poi.getRange()[0], self.poi.getRange()[1]]
        if poi_min is not None:
            poi_range[0] = poi_min
        if poi_max is not None:
            poi_range[1] = poi_max
        self.poi.setRange(poi_range[0], poi_range[1])
        self.stdout.info("INFO: Using POI \"{}\" with range [{}, {}]".format(self.poi.GetName(),
                                                                       poi_range[0],
                                                                       poi_range[1]))
        
        self.model_SB = self.model_base.model_config
        self.poi.setVal(1)
        self.model_SB.SetSnapshot(self.poi)
        
        self.model_B = self.model_base.model_config.Clone("B_only")
        self.poi.setVal(0)
        self.model_B.SetSnapshot(self.poi)
        
        self.seed = seed
        self.random_generator = ROOT.RooRandom.randomGenerator()
        self.random_generator.SetSeed(seed)
        
        pdf_SB = self.model_SB.GetPdf()
        globs_SB = self.model_SB.GetGlobalObservables()
        
        if use_custom_teststat_cls:
            self.load_extension()
            teststat_cls = ROOT.RooStats.ProfileLikelihoodTestStatEx
        else:
            teststat_cls = ROOT.RooStats.ProfileLikelihoodTestStat
        self.plr = teststat_cls(pdf_SB)
        self.plr.SetGlobalObservables(globs_SB)
        self.plr.SetOneSided(True)
        self.plr.SetPrintLevel(print_level)
        self.plr.SetMinimizer(minimizer_type)
        self.plr.SetStrategy(strategy)
        self.plr.SetLOffset(offset)
        self.plr.SetTolerance(tolerance)
        
        self.toy_mc = ROOT.RooStats.ToyMCSampler(self.plr, n_toys)
        
        # If we use the frequentist calculator for counting experiments 
        # (instead of models of distributions) we should instruct the sampler 
        # to generate one event for each toy. ( This is the case because we model 
        # counting experiments in RooFit as a single observation in distribution of event counts. )
        if (not self.model_SB.GetPdf().canBeExtended()):
            self.toy_mc.SetNEventsPerToy(1)        

        if mode == 'toy':
            self.calculator = ROOT.RooStats.FrequentistCalculator(self.model_base.data, 
                                                                  self.model_B, 
                                                                  self.model_SB,
                                                                  self.toy_mc)
        elif mode == 'asymptotic':
            self.calculator = ROOT.RooStats.AsymptoticCalculator(self.model_base.data, 
                                                                 self.model_B, 
                                                                 self.model_SB)
            self.calculator.SetOneSided(True)
            self.calculator.SetQTilde(True)
            self.calculator.SetPrintLevel(print_level)
            self.calculator.Initialize()
        else:
            raise ValueError("mode must be either `toy` or `asymptotic`")

        self.do_coarse_scan = do_coarse_scan
        self.eps            = eps
        self.mode           = mode
        
        self.reset_results()
        

    @semistaticmethod
    def load_extension(self):
        try:
            if not hasattr(ROOT, 'ProfileLikelihoodTestStatEx'):
                result = load_macro('ProfileLikelihoodTestStatEx')
                if hasattr(ROOT, 'ProfileLikelihoodTestStatEx'):
                    self.stdout.info('INFO: Loaded extension module "ProfileLikelihoodTestStatEx"')
        except Exception as e:
            print(e)        
        
    def reset_results(self):
        result = ROOT.RooStats.HypoTestInverterResult("result", self.poi, 0.95)
        result.SetConfidenceLevel(0.95)
        result.UseCLs()
        result.SetInterpolationOption(ROOT.RooStats.HypoTestInverterResult.kLinear)
        self.result = result
        
    @staticmethod
    def scan_result_to_dict(scan_result):
        result = {}
        result["limits"] = {
            "obs": scan_result.UpperLimit(),
            0: scan_result.GetExpectedUpperLimit(0),
            1: scan_result.GetExpectedUpperLimit(1),
            2: scan_result.GetExpectedUpperLimit(2),
            -1: scan_result.GetExpectedUpperLimit(-1),
            -2: scan_result.GetExpectedUpperLimit(-1),
        }
        result["data"] = {
            "mu": [],
            "CLb": [],
            "CLs": [],
            "CLsplusb": [],
            "CLsplusbError": []
        }
        
        for i in range(scan_result.ArraySize()):
            result["data"]["mu"].append(scan_result.GetXValue(i))
            result["data"]["CLb"].append(scan_result.CLb(i))
            result["data"]["CLs"].append(scan_result.CLs(i))
            result["data"]["CLsplusb"].append(scan_result.CLsplusb(i))
            result["data"]["CLsplusbError"].append(scan_result.CLsplusbError(i))
        return result

    def display_limits(self, scan_result):
        limits = ToyLimitCalculator.scan_result_to_dict(scan_result)['limits']
        self.stdout.info("Limit Bands")
        self.stdout.info(f"+2 sigma : {limits[2]}")
        self.stdout.info(f"+1 sigma : {limits[1]}")
        self.stdout.info(f"-1 sigma : {limits[-1]}")
        self.stdout.info(f"-2 sigma : {limits[-2]}")
        self.stdout.info(f"  Median : {limits[0]}")
        self.stdout.info("Observed : {}".format(limits["obs"]))
        
    def check_test_statistics(self, scan_min:float, scan_max:float, steps:int):
        poi_val_0 = self.poi.getVal()
        self.stdout.info("INFO: Checking convergence of test statistics...")
        interval = (scan_max - scan_min) / (steps - 1)
        for i in range(steps):
            mu = scan_min + i * interval
            self.poi.setVal(mu)
            test_statistic = self.plr.EvaluateProfileLikelihood(0, 
                                                                self.model_base.data,
                                                                ROOT.RooArgSet(self.poi))
            self.stdout.info("INFO: mu = {}, test statistic = {}".format(mu, test_statistic))
            if test_statistic < 0:
                raise RuntimeError("test statistic failed to converge")
                
    def append(self, result: Union[ROOT.RooStats.HypoTestInverterResult,
                                   ROOT.RooStats.HypoTestResult]):
        processed_result = merge_toy_results([result], self.poi,
                                             **self.hypotest_options,
                                             silent=True)
        self.result.Add(processed_result)
        
    def run_coarse_scan(self, scan_min:float, scan_max:float):
        
        scale      = 10**(1/3)
        steps      = (math.log(scan_max / scan_min) / math.log(scale)) + 1
        steps_ceil = math.ceil(steps)
        start, end = scan_min, scan_max
        
        self.stdout.info("INFO: Starting coarse search")
        self.stdout.info(f"INFO: Evaluating {steps_ceil} logarithmic points from {start} to {end}")
        
        inverter = ROOT.RooStats.HypoTestInverter(self.calculator)
        inverter.SetVerbose(self.inverter_verbose)
        # 95% CLs limits
        inverter.SetConfidenceLevel(0.95) 
        inverter.UseCLs(True)
        
        self.stdout.info("INFO: Checking for problematic CLs values...")
        
        snapshot_0 = self.freq_calculator.GetNullModel().GetSnapshot()

        for i in range(steps_ceil):
            start = math.exp(math.log(scan_min) + i * math.log(scan_max / scan_min) / (steps - 1))
            self.poi.setVal(start)
            self.freq_calculator.GetNullModel().SetSnapshot(self.poi)
            result = self.freq_calculator.GetHypoTest()
            cls = result.CLs()
            if (math.isfinite(cls) and cls >= 0.):
                break
        if start >= end :
            raise RuntimeError("No acceptable points found in coarse scan")
        for i in range(steps_ceil):
            end = math.exp(math.log(scan_min) + (steps - 1 - i) * math.log(scan_max / scan_min) / (steps - 1))
            self.poi.setVal(end)
            self.freq_calculator.GetNullModel().SetSnapshot(self.poi)
            result = self.freq_calculator.GetHypoTest()
            cls = result.CLs()
            if (math.isfinite(cls) and cls >= 0.):
                break
        if start >= end :
            raise RuntimeError("No acceptable points found in coarse scan")
        steps      = math.log(end / start) / math.log(scale) + 1
        steps_ceil = math.ceil(steps)
        self.freq_calculator.GetNullModel().SetSnapshot(snapshot_0)
        self.stdout.info(f"INFO: Recalculated scan running {steps_ceil} logarithmic points from {start} to {end}")
        
        self.stdout.info("INFO: Running fixed scan")
        inverter.SetFixedScan(steps, start, end, True)
        result = inverter.GetInterval()
        result.SetInterpolationOption(ROOT.RooStats.HypoTestInverterResult.kLinear)
        
        return result        

    def run_scans(self, scan_min:float, scan_max:float, steps:Optional[int]=None, 
                  use_log:bool=False, append:bool=True):
            
        time_start = time.time()
        start, end = scan_min, scan_max
        
        if self.do_coarse_scan:
            coarse_result = self.run_coarse_scan(scan_min, scan_max, mode=mode)
            start = coarse_result.minus_2 / 10**(1/3)
            end   = coarse_result.plus_2 / 10**(1/3)
            if (not math.isfinite(start)) or (not math.isfinite(end)):
                raise RuntimeError("Got non-finite bound from coarse limits")
            if (start < self.poi.getRange()[0]):
                raise RuntimeError("Got lower bound beyond POI minimum range")
            self.stdout.info(f"INFO: Bounds for fine search: [{start}, {end}]")
            self.display_limits(coarse_result)
        
        self.stdout.info("INFO: Starting fine search")
        
        if steps is None:
            scale      = 1. + self.eps
            steps      = (math.log(end / start) / math.log(scale)) + 1
            steps_ceil = math.ceil(steps)
            use_log   = True
            self.stdout.info(f"INFO: Evaluating {steps_ceil} logarithmic points from {start} to {end}")
        else:
            self.stdout.info(f"INFO: Evaluating {steps} points from {start} to {end}")
        inverter = ROOT.RooStats.HypoTestInverter(self.calculator)
        inverter.SetVerbose(self.inverter_verbose)
        # 95% CLs limits
        inverter.SetConfidenceLevel(0.95) 
        inverter.UseCLs(True)
        
        if self.mode == 'toy':
            self.check_test_statistics(start, end, steps)
        
        self.stdout.info("INFO: Running fixed scan")
        inverter.RunFixedScan(steps, start, end, use_log)
        result = inverter.GetInterval()
        result.SetInterpolationOption(ROOT.RooStats.HypoTestInverterResult.kLinear)
        result.SetName("scan_{}_{}_{}".format(str_encode_value(scan_min, 10),
                                              str_encode_value(scan_max, 10),
                                              steps))
        time_end = time.time()
        
        self.stdout.info("INFO: Scan finished. Total time taken: {}s".format(time_end-time_start))
        
        self.display_limits(result)
        
        if append:
            self.append(result)
        
        return result
    
    def run_one_point(self, poi_val:float, append=True):      
        
        poi_val_0 = self.poi.getVal()
        
        self.stdout.info(f"INFO: Running HypoTest on mu value {poi_val}.")
    
        self.poi.setVal(poi_val)
        self.calculator.GetNullModel().SetSnapshot(self.poi)
        result = self.calculator.GetHypoTest()
        result.SetBackgroundAsAlt(True)
        result.SetName(f"{self.poi.GetName()}_{poi_val}")
        
        self.poi.setVal(poi_val_0)
        
        if append and self.mode == 'toy':
            self.append(result)
        
        return result
    
    def get_teststat(self, poi_val:float):
        poi_val_0 = self.poi.getVal()
        self.poi.setVal(poi_val)
        teststat = self.toy_mc.EvaluateTestStatistic(self.model_base.data, 
                                                     ROOT.RooArgSet(self.poi))
        #self.EvaluateProfileLikelihood(0, data, ROOT.RooArgSet(poi)
        self.poi.setVal(poi_val_0)
        return teststat
    

    
    def save_as_root(self, basename:str="toy_result_seed_{seed}.root"):
        if not self.result:
            self.stdout.warning("WARNING: No result to save.")
            return None
        filename = basename.format(seed=self.seed)
        self.write_to_root(self.result, filename)
            
    def write_to_root(self, result, filename):
        f = ROOT.TFile(filename, "RECREATE")
        f.cd()
        result.Write()
        f.Write()
        f.Close()
        self.stdout.info(f"INFO: Saved toy limit result as `{filename}`.")
    
    def save(self, filename:str="toy_limit_result.json"):
        
        if (self.result is None) and (self.coarse_result is None) and (self.one_point_result is None):
            self.stdout.warning("WARNING: No result to save")

        base_dir = os.path.dirname(filename)
        base_name = os.path.basename(filename)
        extension = os.path.splitext(filename)[1]
            
        if self.coarse_result is not None:
            result = self.scan_result_to_dict(self.coarse_result)
            filename_coarse = os.path.join(base_dir, f"{base_name}_coarse{extension}")
            with open(filename_coarse) as f:
                json.dump(result, f, indent=2)
            self.stdout.info(f"INFO: Saved coarse toy limit result as `{filename}`.")
            
        if self.result is not None:
            result = self.scan_result_to_dict(self.result)
            with open(filename) as f:
                json.dump(result, f, indent=2)
            self.stdout.info(f"INFO: Saved toy limit result as `{filename}`.")

        if self.one_point_result is not None:
            for mu_val in self.one_point_result:
                mu_str = str_encode_value(mu_val)
                result = self.scan_result_to_dict(self.one_point_result[mu_val])
                filename_one_point = os.path.join(base_dir, f"{base_name}_{mu_str}{extension}")
                with open(filename_one_point) as f:
                    json.dump(result, f, indent=2)
                self.stdout.info(f"INFO: Saved toy limit result for mu = {mu_val} as `{filename}`.")
                
def evaluate_toy_limits(filename:str, data_name:str, 
                        scan_min:Optional[float]=None, 
                        scan_max:Optional[float]=None, 
                        steps:int=10,
                        mu_val:Optional[float]=None,
                        poi_min:Optional[float]=None, poi_max:Optional[float]=None,
                        n_toys:int=1000, seed:int=1234, 
                        outname:str="toy_result_seed_{seed}.root",
                        poi_name:Optional[str]=None,
                        minimizer_type:str="Minuit2", strategy:int=1,
                        eps:float=0.05, tolerance:float=1., offset:bool=True,
                        print_level:int=-1, do_coarse_scan:bool=False, 
                        fix_param:str='', profile_param:str='',
                        snapshot_name:Optional[Union[List[str], str]]=None,                 
                        verbosity:Optional[Union[int, str]]=None):
    toy_limit_calculator = ToyLimitCalculator(filename=filename,
                                              data_name=data_name,
                                              n_toys=n_toys, seed=seed,
                                              poi_name=poi_name,
                                              poi_min=poi_min,
                                              poi_max=poi_max,
                                              minimizer_type=minimizer_type,
                                              strategy=strategy,
                                              eps=eps, tolerance=tolerance,
                                              offset=offset,
                                              print_level=print_level,
                                              do_coarse_scan=do_coarse_scan,
                                              fix_param=fix_param,
                                              profile_param=profile_param,
                                              snapshot_name=snapshot_name,
                                              verbosity=verbosity)
    if all([p is not None for p in [scan_min, scan_max]]) and (mu_val is None):
        toy_limit_calculator.run_scans(scan_min=scan_min, scan_max=scan_max, steps=steps)
    elif (mu_val is not None) and all([p is None for p in [scan_min, scan_max]]):
        toy_limit_calculator.run_one_point(mu_val)
    else:
        raise ValueError("please provide either (scan_min, scan_max, steps) for running a scan or (mu_val)"
                         " for running a single point")
    toy_limit_calculator.save_as_root(basename=outname)
                
def evaluate_batched_toy_limits(filename:str, data_name:str, 
                                scan_min:Optional[float]=None,
                                scan_max:Optional[float]=None,
                                steps:int=10, 
                                mu_val:Optional[float]=None,
                                poi_min:Optional[float]=None, poi_max:Optional[float]=None,
                                n_toys:int=1000, batchsize:int=100, seed:int=1234, 
                                outname:str="toy_result_seed_{seed}_batch_{batch}.root",
                                poi_name:Optional[str]=None, minimizer_type:str="Minuit2", 
                                strategy:int=1, eps:float=0.05, tolerance:float=1., 
                                offset:bool=True, print_level:int=-1,
                                do_coarse_scan:bool=False, parallel:int=-1,
                                fix_param:str='', profile_param:str='',
                                snapshot_name:Optional[Union[List[str], str]]=None,                 
                                verbosity:Optional[Union[int, str]]=None):
    if n_toys % batchsize == 0:
        toy_batches = [batchsize] * (n_toys // batchsize)
    else:
        toy_batches = [batchsize] * (n_toys // batchsize) + [n_toys % batchsize]
    seeds = [seed + 10000*i for i in range(len(toy_batches))]
    
    dirname = os.path.dirname(outname)
    if dirname and (not os.path.exists(dirname)):
        os.makedirs(dirname, exist_ok=True)
    
    outnames = [outname.format(seed=seed, batch=i) for i in range(len(toy_batches))]
    
    args = (repeat(filename), repeat(data_name), repeat(scan_min), repeat(scan_max),
            repeat(steps), repeat(mu_val), repeat(poi_min), repeat(poi_max),
            toy_batches, seeds, outnames, repeat(poi_name), repeat(minimizer_type), 
            repeat(strategy), repeat(eps), repeat(tolerance), repeat(offset),
            repeat(print_level), repeat(do_coarse_scan), repeat(fix_param), 
            repeat(profile_param), repeat(snapshot_name), repeat(verbosity))
    execute_multi_tasks(evaluate_toy_limits, *args, parallel=parallel)


from typing import Optional, Union, List, Dict

import quickstats

from quickstats.utils.io import VerbosePrint

class AbstractObject(object):
    
    stdout = quickstats._PRINT_
    
    def __init__(self, verbosity:Optional[Union[int, str]]=None):
        
        if verbosity is None:
            self.stdout = quickstats._PRINT_
        else:
            self.stdout = VerbosePrint(verbosity)
        
            
    def set_verbosity(self, verbosity:Optional[Union[int, str]]):
        """
            Change the verbosity of the current class. This will detach the class's standard output 
            from the centrally-managed standard output.
        """
        self.stdout = VerbosePrint(verbosity)


BUILTIN_METHODS = \
{
    "Namespace_ROOT_VecOps": "using namespace ROOT::VecOps;", 
    "Namespace_ROOT_Math": "using namespace ROOT::Math;",
    "SizeOf" : "template<typename T>\n"
               "auto SizeOf(T vec)\n"
               "{return vec.size();}",
    "CosineThetaStar": "template<typename T1, typename T2>\n"
                       "auto CosineThetaStar(T1 P1, T2 P2){\n"
                       "auto P12 = P1 + P2;\n"
                       "double P1_plus  = (P1.E() + P1.Pz())/1.41421356;\n"
                       "double P2_plus  = (P2.E() + P2.Pz())/1.41421356;\n"
                       "double P1_minus = (P1.E() - P1.Pz())/1.41421356;\n"
                       "double P2_minus = (P2.E() - P2.Pz())/1.41421356;\n"
                       "double M12_squared = pow(P12.M(),2);\n"
                       "double PT12_squared = pow(P12.Pt(),2);\n"
                       "return 2*((P1_plus*P2_minus -P1_minus*P2_plus)/"
                       "sqrt(M12_squared*(M12_squared+PT12_squared)))*(P12.Pz()/abs(P12.Pz()));}\n"
}


from typing import Optional, Union, Dict, List

from quickstats.plots.template import single_frame, parse_styles, parse_analysis_label_options
from quickstats.utils.common_utils import combine_dict

class AbstractPlot:
    
    COLOR_PALLETE = {}
    COLOR_PALLETE_SEC = {}
    
    def __init__(self,
                 color_pallete:Optional[Dict]=None,
                 color_pallete_sec:Optional[Dict]=None,
                 styles:Optional[Union[Dict, str]]=None,
                 analysis_label_options:Optional[Dict]=None):
        
        self.color_pallete = combine_dict(self.COLOR_PALLETE, color_pallete)
        self.color_pallete_sec = combine_dict(self.COLOR_PALLETE_SEC, color_pallete_sec)
            
        self.styles = parse_styles(styles)
        
        if analysis_label_options is None:
            self.analysis_label_options = None
        else:
            self.analysis_label_options = parse_analysis_label_options(analysis_label_options)
            
        self.legend_order = self.get_default_legend_order()
        
        self.legend_data = {}
        self.legend_data_sec = {}
    
    def get_default_legend_order(self):
        return []
    
    def update_legend_handles(self, handles:Dict, sec:bool=False):
        if not sec:
            legend_data = self.legend_data
        else:
            legend_data = self.legend_data_sec
            
        for key in handles:
            handle = handles[key]
            if isinstance(handle, (tuple, list)):
                label = handle[0].get_label()
            else:
                label = handle.get_label()
            if label and not label.startswith('_'):
                legend_data[key] = {
                    'handle': handle,
                    'label': label
                }
            else:
                raise RuntimeError(f"the handle {handle} does not have an associated label")

    def get_legend_handles_labels(self, sec:bool=False):
        handles = []
        labels = []
        if not sec:
            legend_data = self.legend_data
        else:
            legend_data = self.legend_data_sec        
        for key in self.legend_order:
            if key in legend_data:
                handle = legend_data[key]['handle']
                label = legend_data[key]['label']
                handles.append(handle)
                labels.append(label)
        return handles, labels
    
    def draw_frame(self, **kwargs):
        ax = single_frame(styles=self.styles,
                          analysis_label_options=self.analysis_label_options,
                          **kwargs)
        return ax
    
    def draw_axis_labels(self, ax, xlabel:Optional[str]=None, ylabel:Optional[str]=None):
        if xlabel is not None:
            ax.set_xlabel(xlabel, **self.styles['xlabel'])
        if ylabel is not None:
            ax.set_ylabel(ylabel, **self.styles['ylabel'])